"""Base classes for all kernels."""

from __future__ import annotations

import gc
from abc import ABC
from typing import TYPE_CHECKING, Any

from attrs import define

from baybe.exceptions import UnmatchedAttributeError
from baybe.priors.base import Prior
from baybe.serialization.core import (
    converter,
    get_base_structure_hook,
    unstructure_base,
)
from baybe.serialization.mixin import SerialMixin
from baybe.utils.basic import get_baseclasses, match_attributes

if TYPE_CHECKING:
    import torch

    from baybe.surrogates.gaussian_process.kernel_factory import PlainKernelFactory


@define(frozen=True)
class Kernel(ABC, SerialMixin):
    """Abstract base class for all kernels."""

    def to_factory(self) -> PlainKernelFactory:
        """Wrap the kernel in a :class:`baybe.surrogates.gaussian_process.kernel_factory.PlainKernelFactory`."""  # noqa: E501
        from baybe.surrogates.gaussian_process.kernel_factory import PlainKernelFactory

        return PlainKernelFactory(self)

    def to_gpytorch(
        self,
        *,
        ard_num_dims: int | None = None,
        batch_shape: torch.Size | None = None,
        active_dims: tuple[int, ...] | None = None,
    ):
        """Create the gpytorch representation of the kernel."""
        import gpytorch.kernels

        # Extract keywords with non-default values. This is required since gpytorch
        # makes use of kwargs, i.e. differentiates if certain keywords are explicitly
        # passed or not. For instance, `ard_num_dims = kwargs.get("ard_num_dims", 1)`
        # fails if we explicitly pass `ard_num_dims=None`.
        kw: dict[str, Any] = dict(
            ard_num_dims=ard_num_dims, batch_shape=batch_shape, active_dims=active_dims
        )
        kw = {k: v for k, v in kw.items() if v is not None}

        # Get corresponding gpytorch kernel class and its base classes
        kernel_cls = getattr(gpytorch.kernels, self.__class__.__name__)
        base_classes = get_baseclasses(kernel_cls, abstract=True)

        # Fetch the necessary gpytorch constructor parameters of the kernel.
        # NOTE: In gpytorch, some attributes (like the kernel lengthscale) are handled
        #   via the `gpytorch.kernels.Kernel` base class. Hence, it is not sufficient to
        #   just check the fields of the actual class, but also those of its base
        #   classes.
        kernel_attrs: dict[str, Any] = {}
        unmatched_attrs: dict[str, Any] = {}
        for cls in [kernel_cls, *base_classes]:
            matched, unmatched = match_attributes(self, cls.__init__, strict=False)
            kernel_attrs.update(matched)
            unmatched_attrs.update(unmatched)

        # Sanity check: all attributes of the BayBE kernel need a corresponding match
        # in the gpytorch kernel (otherwise, the BayBE kernel class is misconfigured).
        # Exception: initial values are not used during construction but are set
        # on the created object (see code at the end of the method).
        missing = set(unmatched) - set(kernel_attrs)
        if leftover := {m for m in missing if not m.endswith("_initial_value")}:
            raise UnmatchedAttributeError(leftover)

        # Convert specified priors to gpytorch, if provided
        prior_dict = {
            key: value.to_gpytorch()
            for key, value in kernel_attrs.items()
            if isinstance(value, Prior)
        }

        # Convert specified inner kernels to gpytorch, if provided
        kernel_dict = {
            key: value.to_gpytorch(**kw)
            for key, value in kernel_attrs.items()
            if isinstance(value, Kernel)
        }

        # Create the kernel with all its inner gpytorch objects
        kernel_attrs.update(kernel_dict)
        kernel_attrs.update(prior_dict)
        gpytorch_kernel = kernel_cls(**kernel_attrs, **kw)

        # If the kernel has a lengthscale, set its initial value
        if kernel_cls.has_lengthscale:
            import torch

            from baybe.utils.torch import DTypeFloatTorch

            # We can ignore mypy here and simply assume that the corresponding BayBE
            # kernel class has the necessary lengthscale attribute defined. This is
            # safer than using a `hasattr` check in the above if-condition since for
            # the latter the code would silently fail when forgetting to add the
            # attribute to a new kernel class / misspelling it.
            if (initial_value := self.lengthscale_initial_value) is not None:  # type: ignore[attr-defined]
                gpytorch_kernel.lengthscale = torch.tensor(
                    initial_value, dtype=DTypeFloatTorch
                )

        return gpytorch_kernel


@define(frozen=True)
class BasicKernel(Kernel, ABC):
    """Abstract base class for all basic kernels."""


@define(frozen=True)
class CompositeKernel(Kernel, ABC):
    """Abstract base class for all composite kernels."""


# Register (un-)structure hooks
converter.register_structure_hook(Kernel, get_base_structure_hook(Kernel))
converter.register_unstructure_hook(Kernel, unstructure_base)

# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Direct arylation reaction benchmark (non-transfer learning)."""

from __future__ import annotations

import pandas as pd

from baybe.campaign import Campaign
from baybe.objectives import SingleTargetObjective
from baybe.parameters import (
    CategoricalParameter,
    NumericalDiscreteParameter,
    SubstanceParameter,
)
from baybe.recommenders import RandomRecommender
from baybe.searchspace import SearchSpace
from baybe.simulation import simulate_scenarios
from baybe.targets import NumericalTarget
from benchmarks.data.utils import DATA_PATH
from benchmarks.definition import (
    ConvergenceBenchmark,
    ConvergenceBenchmarkSettings,
)


def direct_arylation_multi_batch(
    settings: ConvergenceBenchmarkSettings,
) -> pd.DataFrame:
    """Benchmark function for direct arylation reaction."""
    data = pd.read_table(
        DATA_PATH / "direct_arylation" / "data.csv",
        sep=",",
        index_col=0,
    )
    non_substance_paramters = [
        NumericalDiscreteParameter(
            name="Concentration",
            values=sorted(data["Concentration"].unique()),
        ),
        NumericalDiscreteParameter(
            name="Temp_C",
            values=sorted(data["Temp_C"].unique()),
        ),
    ]
    catgorical_parameters = [
        CategoricalParameter(name=substance, values=data[substance].unique())
        for substance in ["Solvent", "Base", "Ligand"]
    ]
    rdkit_parameters = [
        SubstanceParameter(
            name=substance,
            data=dict(zip(data[substance], data[f"{substance}_SMILES"])),
            encoding="RDKIT2DDESCRIPTORS",
        )
        for substance in ["Solvent", "Base", "Ligand"]
    ]
    rdkit_parameters = [
        SubstanceParameter(
            name=substance,
            data=dict(zip(data[substance], data[f"{substance}_SMILES"])),
            encoding="MORDRED",
        )
        for substance in ["Solvent", "Base", "Ligand"]
    ]

    categorical_searchspace = SearchSpace.from_product(
        parameters=[*catgorical_parameters, *non_substance_paramters]
    )
    rdkit_searchspace = SearchSpace.from_product(
        parameters=[*rdkit_parameters, *non_substance_paramters]
    )
    mordred_searchspace = SearchSpace.from_product(
        parameters=[*rdkit_parameters, *non_substance_paramters]
    )
    target = NumericalTarget(name="yield", mode="MAX")
    objective = SingleTargetObjective(target=target)

    scenarios: dict[str, Campaign] = {
        "Random Recommender": Campaign(
            searchspace=categorical_searchspace,
            recommender=RandomRecommender(),
            objective=objective,
        ),
        "Categorical": Campaign(
            searchspace=categorical_searchspace,
            objective=objective,
        ),
        "RDKIT": Campaign(
            searchspace=rdkit_searchspace,
            objective=objective,
        ),
        "Mordred": Campaign(
            searchspace=mordred_searchspace,
            objective=objective,
        ),
    }

    results_df = simulate_scenarios(
        scenarios,
        data,
        batch_size=settings.batch_size,
        n_doe_iterations=settings.n_doe_iterations,
        n_mc_iterations=settings.n_mc_iterations,
        impute_mode="error",
    )
    return results_df


benchmark_config = ConvergenceBenchmarkSettings(
    batch_size=3,
    n_doe_iterations=30,
    n_mc_iterations=100,
)

direct_arylation_multi_batch_benchmark = ConvergenceBenchmark(
    function=direct_arylation_multi_batch,
    optimal_target_values={"yield": 100},
    settings=benchmark_config,
)



"""Validation functionality for objectives."""

from collections.abc import Collection

from baybe.targets.base import Target


def validate_target_names(  # noqa: DOC101, DOC103
    _, __, targets: Collection[Target]
) -> None:
    """An attrs-compatible validator to assert unique target names.

    Raises:
        ValueError: If the given collection contains targets with the same name.
    """  # noqa: D401
    if len(names := [t.name for t in targets]) != len(set(names)):
        raise ValueError("All targets must have unique names.")



"""Optional polars import."""

from baybe.exceptions import OptionalImportError

try:
    import polars
except ModuleNotFoundError as ex:
    raise OptionalImportError(name="polars", group="polars") from ex

__all__ = [
    "polars",
]



"""Functionality for multi-target objectives."""

import gc
import warnings
from typing import ClassVar

import pandas as pd
from attrs import define, field
from attrs.validators import deep_iterable, instance_of, min_len
from typing_extensions import override

from baybe.objectives.base import Objective
from baybe.objectives.validation import validate_target_names
from baybe.targets import NumericalTarget, TargetMode
from baybe.targets.base import Target
from baybe.utils.basic import to_tuple
from baybe.utils.dataframe import transform_target_columns


def _block_minmax_transforms(_, __, target: Target) -> None:  # noqa: DOC101, DOC103
    """An attrs-compatible validator to assert that a target has no transform.

    Raises:
        ValueError: If the target has a transform.
    """  # noqa: D401
    if (
        isinstance(target, NumericalTarget)
        and target.mode
        in [
            TargetMode.MIN,
            TargetMode.MAX,
        ]
        and target.transformation is not None
    ):
        raise ValueError(
            f"'{ParetoObjective.__name__}' does not support transforms for targets "
            f"with modes '{TargetMode.MIN.name}' or '{TargetMode.MAX.name}'. "
            f"Please use untransformed/unbounded targets in this case."
        )


@define(frozen=True, slots=False)
class ParetoObjective(Objective):
    """An objective handling multiple targets in a Pareto sense."""

    is_multi_output: ClassVar[bool] = True
    # See base class.

    _targets: tuple[Target, ...] = field(
        converter=to_tuple,
        validator=[
            min_len(2),
            deep_iterable(
                member_validator=(instance_of(Target), _block_minmax_transforms)
            ),
            validate_target_names,
        ],
        alias="targets",
    )
    "The targets considered by the objective."

    @override
    @property
    def targets(self) -> tuple[Target, ...]:
        return self._targets

    @override
    @property
    def n_outputs(self) -> int:
        return len(self._targets)

    @override
    def transform(
        self,
        df: pd.DataFrame | None = None,
        /,
        *,
        allow_missing: bool = False,
        allow_extra: bool | None = None,
        data: pd.DataFrame | None = None,
    ) -> pd.DataFrame:
        # >>>>>>>>>> Deprecation
        if not ((df is None) ^ (data is None)):
            raise ValueError(
                "Provide the dataframe to be transformed as first positional argument."
            )

        if data is not None:
            df = data
            warnings.warn(
                "Providing the dataframe via the `data` argument is deprecated and "
                "will be removed in a future version. Please pass your dataframe "
                "as positional argument instead.",
                DeprecationWarning,
            )

        # Mypy does not infer from the above that `df` must be a dataframe here
        assert isinstance(df, pd.DataFrame)

        if allow_extra is None:
            allow_extra = True
            if set(df.columns) - {p.name for p in self.targets}:
                warnings.warn(
                    "For backward compatibility, the new `allow_extra` flag is set "
                    "to `True` when left unspecified. However, this behavior will be "
                    "changed in a future version. If you want to invoke the old "
                    "behavior, please explicitly set `allow_extra=True`.",
                    DeprecationWarning,
                )
        # <<<<<<<<<< Deprecation

        return transform_target_columns(
            df, self.targets, allow_missing=allow_missing, allow_extra=allow_extra
        )


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Information about availability of optional dependencies."""

import os
import sys
from contextlib import contextmanager
from importlib.util import find_spec


@contextmanager
def exclude_sys_path(path: str, /):  # noqa: DOC402, DOC404
    """Temporarily remove a specified path from `sys.path`.

    Args:
        path: The path to exclude from the search.
    """
    original_sys_path = sys.path[:]
    if path in sys.path:
        sys.path.remove(path)
    try:
        yield
    finally:
        sys.path = original_sys_path


# Individual packages
with exclude_sys_path(os.getcwd()):
    FLAKE8_INSTALLED = find_spec("flake8") is not None
    LIME_INSTALLED = find_spec("lime") is not None
    ONNX_INSTALLED = find_spec("onnxruntime") is not None
    NGBOOST_INSTALLED = find_spec("ngboost") is not None
    POLARS_INSTALLED = find_spec("polars") is not None
    PRE_COMMIT_INSTALLED = find_spec("pre_commit") is not None
    PYDOCLINT_INSTALLED = find_spec("pydoclint") is not None
    RUFF_INSTALLED = find_spec("ruff") is not None
    SHAP_INSTALLED = find_spec("shap") is not None
    SKFP_INSTALLED = find_spec("skfp") is not None  # scikit-fingerprints
    STREAMLIT_INSTALLED = find_spec("streamlit") is not None
    XYZPY_INSTALLED = find_spec("xyzpy") is not None

# TODO: `find_spec` does not work for `typos` (probably because it's a rust package),
#   neither seem any versions of `pip show typos` / `pip freeze | grep typos` /
#   `any(pkg.name == "typos" for pkg in pkgutils.iter_modules())` called as a subprocess
#   work in CI (probably because stuff is running inside tox and/or paths need to be
#   configured properly). As a temporary workaround – since the test collection does not
#   directly depend on the flag – we thus simply set it to `True`.
TYPOS_INSTALLED = True

# Information on whether all required packages for certain functionality are available
CHEM_INSTALLED = SKFP_INSTALLED
INSIGHTS_INSTALLED = SHAP_INSTALLED and LIME_INSTALLED
LINT_INSTALLED = all(
    (
        FLAKE8_INSTALLED,
        PRE_COMMIT_INSTALLED,
        PYDOCLINT_INSTALLED,
        RUFF_INSTALLED,
        TYPOS_INSTALLED,
    )
)



"""Benchmark domains."""

from benchmarks.definition.base import Benchmark
from benchmarks.domains.direct_arylation.direct_arylation_multi_batch import (
    direct_arylation_multi_batch_benchmark,
)
from benchmarks.domains.direct_arylation.direct_arylation_single_batch import (
    direct_arylation_single_batch_benchmark,
)
from benchmarks.domains.hartmann.hartmann_3d import hartmann_3d_benchmark
from benchmarks.domains.hartmann.hartmann_3d_discretized import (
    hartmann_3d_discretized_benchmark,
)
from benchmarks.domains.hartmann.hartmann_6d import hartmann_6d_benchmark
from benchmarks.domains.synthetic_2C1D_1C import synthetic_2C1D_1C_benchmark
from benchmarks.domains.transfer_learning.aryl_halides.sou_CT_I_tar_BM import (
    aryl_halide_CT_I_BM_tl_benchmark,
)
from benchmarks.domains.transfer_learning.aryl_halides.sou_CT_tar_IM import (
    aryl_halide_CT_IM_tl_benchmark,
)
from benchmarks.domains.transfer_learning.aryl_halides.sout_IP_tar_CP import (
    aryl_halide_IP_CP_tl_benchmark,
)
from benchmarks.domains.transfer_learning.direct_arylation.temperature_tl import (
    direct_arylation_tl_temperature_benchmark,
)
from benchmarks.domains.transfer_learning.easom.easom_tl_47_negate_noise5 import (
    easom_tl_47_negate_noise5_benchmark,
)
from benchmarks.domains.transfer_learning.hartmann.hartmann_tl_3_20_15 import (
    hartmann_tl_3_20_15_benchmark,
)
from benchmarks.domains.transfer_learning.michalewicz.michalewicz_tl_continuous import (
    michalewicz_tl_continuous_benchmark,
)

BENCHMARKS: list[Benchmark] = [
    direct_arylation_multi_batch_benchmark,
    direct_arylation_single_batch_benchmark,
    hartmann_3d_discretized_benchmark,
    synthetic_2C1D_1C_benchmark,
    hartmann_3d_benchmark,
    hartmann_6d_benchmark,
    aryl_halide_CT_IM_tl_benchmark,
    aryl_halide_IP_CP_tl_benchmark,
    aryl_halide_CT_I_BM_tl_benchmark,
    direct_arylation_tl_temperature_benchmark,
    easom_tl_47_negate_noise5_benchmark,
    hartmann_tl_3_20_15_benchmark,
    michalewicz_tl_continuous_benchmark,
]


__all__ = ["BENCHMARKS"]



## Modeling a Mixture in Traditional Representation

# When modeling mixtures, we are often faced with a large set of ingredients to choose
# from. A common way to formalize this type of selection problem is to assign each
# ingredient its own numerical parameter representing the amount of the ingredient in
# the mixture. A sum constraint imposed on all parameters then ensures that the total
# amount of ingredients in the mix is always 100%. In addition, there could be other
# constraints, for instance, to impose further restrictions on individual subgroups of
# ingredients. In BayBE's language, we call this the *traditional mixture
# representation*.

# In this example, we demonstrate how to create a search space in this representation,
# using a simple mixture of up to six components, which are divided into three
# subgroups: solvents, bases and phase agents.

# ```{admonition} Slot-based Representation
# :class: seealso
# For an alternative way to describe mixtures, see our
# [slot-based representation](/examples/Mixtures/slot_based.md).
# ```

### Imports

import numpy as np
import pandas as pd

from baybe.constraints import ContinuousLinearConstraint
from baybe.parameters import NumericalContinuousParameter
from baybe.recommenders import RandomRecommender
from baybe.searchspace import SearchSpace

### Parameter Setup

# We start by creating lists containing our substance labels according to their
# subgroups:

g1 = ["Solvent1", "Solvent2"]
g2 = ["Base1", "Base2"]
g3 = ["PhaseAgent1", "PhaseAgent2"]

# Next, we create continuous parameters describing the substance amounts for each group.
# Here, the maximum amount for each substance depends on its group, i.e. we allow
# adding more of a solvent compared to a base or a phase agent:

p_g1_amounts = [
    NumericalContinuousParameter(name=f"{name}", bounds=(0, 80)) for name in g1
]
p_g2_amounts = [
    NumericalContinuousParameter(name=f"{name}", bounds=(0, 20)) for name in g2
]
p_g3_amounts = [
    NumericalContinuousParameter(name=f"{name}", bounds=(0, 5)) for name in g3
]

### Constraints Setup

# Now, we set up our constraints. We start with the overall mixture constraint, ensuring
# the total of all ingredients is 100%:

c_total_sum = ContinuousLinearConstraint(
    parameters=g1 + g2 + g3,
    operator="=",
    coefficients=[1] * len(g1 + g2 + g3),
    rhs=100,
)

# Additionally, we require bases make up at least 10% of the mixture:

c_g2_min = ContinuousLinearConstraint(
    parameters=g2,
    operator=">=",
    coefficients=[1] * len(g2),
    rhs=10,
)

# By contrast, phase agents should make up no more than 5%:

c_g3_max = ContinuousLinearConstraint(
    parameters=g3,
    operator="<=",
    coefficients=[1] * len(g3),
    rhs=5,
)

### Search Space Creation

# Having both parameter and constraint definitions at hand, we can create our
# search space:

searchspace = SearchSpace.from_product(
    parameters=[*p_g1_amounts, *p_g2_amounts, *p_g3_amounts],
    constraints=[c_total_sum, c_g2_min, c_g3_max],
)


### Verification of Constraints

# To verify that the constraints imposed above are fulfilled, let us draw some
# random points from the search space:

recommendations = RandomRecommender().recommend(batch_size=10, searchspace=searchspace)
print(recommendations)

# Computing the respective row sums reveals the expected result:

stats = pd.DataFrame(
    {
        "Total": recommendations.sum(axis=1),
        "Total_Bases": recommendations[g2].sum(axis=1),
        "Total_Phase_Agents": recommendations[g3].sum(axis=1),
    }
)
print(stats)

assert np.allclose(stats["Total"], 100)
assert (stats["Total_Bases"] >= 10).all()
assert (stats["Total_Phase_Agents"] <= 5).all()



"""Base classes for all objectives."""

import gc
from abc import ABC, abstractmethod
from typing import ClassVar

import cattrs
import pandas as pd
from attrs import define

from baybe.serialization.core import (
    converter,
    get_base_structure_hook,
    unstructure_base,
)
from baybe.serialization.mixin import SerialMixin
from baybe.targets.base import Target

# TODO: Reactive slots in all classes once cached_property is supported:
#   https://github.com/python-attrs/attrs/issues/164


@define(frozen=True, slots=False)
class Objective(ABC, SerialMixin):
    """Abstract base class for all objectives."""

    is_multi_output: ClassVar[bool]
    """Class variable indicating if the objective produces multiple outputs."""

    @property
    @abstractmethod
    def targets(self) -> tuple[Target, ...]:
        """The targets included in the objective."""

    @property
    @abstractmethod
    def n_outputs(self) -> int:
        """The number of outputs of the objective."""

    @abstractmethod
    def transform(
        self,
        df: pd.DataFrame,
        /,
        *,
        allow_missing: bool = False,
        allow_extra: bool = False,
    ) -> pd.DataFrame:
        """Transform target values from experimental to computational representation.

        Args:
            df: The dataframe to be transformed. The allowed columns of the dataframe
                are dictated by the ``allow_missing`` and ``allow_extra`` flags.
            allow_missing: If ``False``, each target of the objective must have
                exactly one corresponding column in the given dataframe. If ``True``,
                the dataframe may contain only a subset of target columns.
            allow_extra: If ``False``, each column present in the dataframe must
                correspond to exactly one target of the objective. If ``True``, the
                dataframe may contain additional non-target-related columns, which
                will be ignored.

        Returns:
            A corresponding dataframe with the targets in computational representation.
        """


def to_objective(x: Target | Objective, /) -> Objective:
    """Convert a target into an objective (with objective passthrough)."""
    return x if isinstance(x, Objective) else x.to_objective()


# Register (un-)structure hooks
converter.register_structure_hook(
    Objective,
    get_base_structure_hook(
        Objective,
        overrides=dict(
            _target=cattrs.override(rename="target"),
            _targets=cattrs.override(rename="targets"),
        ),
    ),
)
converter.register_unstructure_hook(
    Objective,
    lambda x: unstructure_base(
        x,
        overrides=dict(
            _target=cattrs.override(rename="target"),
            _targets=cattrs.override(rename="targets"),
        ),
    ),
)
# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Benchmark on Direct Arylation data for transfer learning.

This benchmark uses one temperature as source and another temperature as target.
"""

from __future__ import annotations

import pandas as pd

from baybe.campaign import Campaign
from baybe.objectives import SingleTargetObjective
from baybe.parameters import (
    NumericalDiscreteParameter,
    SubstanceParameter,
    TaskParameter,
)
from baybe.parameters.base import DiscreteParameter
from baybe.searchspace import SearchSpace
from baybe.simulation import simulate_scenarios
from baybe.targets import NumericalTarget
from benchmarks.data.utils import DATA_PATH
from benchmarks.definition import (
    ConvergenceBenchmarkSettings,
)
from benchmarks.definition.convergence import ConvergenceBenchmark


def load_data() -> pd.DataFrame:
    """Load data for benchmark."""
    data = pd.read_table(
        DATA_PATH / "direct_arylation" / "data.csv",
        sep=",",
        index_col=0,
        dtype={"Temp_C": str},
    )
    return data


def make_searchspace(
    data: pd.DataFrame,
    use_task_parameter: bool,
) -> SearchSpace:
    """Create the search space for the benchmark."""
    params: list[DiscreteParameter] = [
        SubstanceParameter(
            name=substance,
            data=dict(zip(data[substance], data[f"{substance}_SMILES"])),
            encoding="RDKIT2DDESCRIPTORS",
        )
        for substance in ["Solvent", "Base", "Ligand"]
    ] + [
        NumericalDiscreteParameter(
            name="Concentration",
            values=sorted(data["Concentration"].unique()),
        ),
    ]
    if use_task_parameter:
        params.append(
            TaskParameter(
                name="Temp_C",
                values=["90", "105", "120"],
                active_values=["105"],
            )
        )
    return SearchSpace.from_product(parameters=params)


def make_objective() -> SingleTargetObjective:
    """Create the objective for the benchmark."""
    return SingleTargetObjective(NumericalTarget(name="yield", mode="MAX"))


def make_lookup(data: pd.DataFrame) -> pd.DataFrame:
    """Create the lookup for the benchmark.

    Note that we filter the data to only include the target tasks.
    Without the filtering, there would be multiple entries for the same parameter
    configuration. Since this might yield issues for the non-transfer learning
    campaigns, we filter the data to only include the target tasks.
    """
    return data[data["Temp_C"] == "105"]


def make_initial_data(data: pd.DataFrame) -> pd.DataFrame:
    """Create the initial data for the benchmark."""
    return data[data["Temp_C"] != "105"]


def direct_arylation_tl_temperature(
    settings: ConvergenceBenchmarkSettings,
) -> pd.DataFrame:
    """Benchmark on Direct Arylation data for transfer learning.

    Key characteristics:
    • Uses one temperature as source (90°C and 120°C) and another as target (105°C)
    • Compares transfer learning vs. non-transfer learning approaches
    • Tests varying amounts of source data:
      - 1% of source data
      - 10% of source data
      - 20% of source data
    • Includes baseline with no transfer learning (0% source data)
    • Parameters:
      - Solvent: Substance with RDKIT2DDESCRIPTORS encoding
      - Base: Substance with RDKIT2DDESCRIPTORS encoding
      - Ligand: Substance with RDKIT2DDESCRIPTORS encoding
      - Concentration: Numerical discrete
      - Temp_C: Task parameter (90°C, 105°C, 120°C)
    • Target: Reaction yield (continuous)
    • Objective: Maximization

    Args:
        settings: Configuration settings for the convergence benchmark

    Returns:
        DataFrame containing benchmark results for all test cases
    """
    data = load_data()

    searchspace = make_searchspace(
        data=data,
        use_task_parameter=True,
    )
    searchspace_nontl = make_searchspace(
        data=data,
        use_task_parameter=False,
    )

    lookup = make_lookup(data)
    initial_data = make_initial_data(data)
    objective = make_objective()

    tl_campaign = Campaign(searchspace=searchspace, objective=objective)
    non_tl_campaign = Campaign(searchspace=searchspace_nontl, objective=objective)

    results = []
    for p in [0.01, 0.1, 0.2]:
        results.append(
            simulate_scenarios(
                {
                    f"{int(100 * p)}": tl_campaign,
                    f"{int(100 * p)}_naive": non_tl_campaign,
                },
                lookup,
                initial_data=[
                    initial_data.sample(frac=p) for _ in range(settings.n_mc_iterations)
                ],
                batch_size=settings.batch_size,
                n_doe_iterations=settings.n_doe_iterations,
                impute_mode="error",
            )
        )
    results.append(
        simulate_scenarios(
            {"0": tl_campaign, "0_naive": non_tl_campaign},
            lookup,
            batch_size=settings.batch_size,
            n_doe_iterations=settings.n_doe_iterations,
            n_mc_iterations=settings.n_mc_iterations,
            impute_mode="error",
        )
    )
    return pd.concat(results)


benchmark_config = ConvergenceBenchmarkSettings(
    batch_size=2,
    n_doe_iterations=20,
    n_mc_iterations=55,
)

direct_arylation_tl_temperature_benchmark = ConvergenceBenchmark(
    function=direct_arylation_tl_temperature,
    optimal_target_values={"yield": 100},
    settings=benchmark_config,
)



"""Discretized version of the Hartmann function in 3 dimensions as a benchmark."""

from __future__ import annotations

import numpy as np
from botorch.test_functions.synthetic import Hartmann
from pandas import DataFrame

from baybe.campaign import Campaign
from baybe.parameters import NumericalDiscreteParameter
from baybe.recommenders import RandomRecommender
from baybe.searchspace import SearchSpace
from baybe.simulation import simulate_scenarios
from baybe.targets import NumericalTarget
from baybe.utils.dataframe import arrays_to_dataframes
from benchmarks.definition.convergence import (
    ConvergenceBenchmark,
    ConvergenceBenchmarkSettings,
)


def hartmann_3d_discretized(settings: ConvergenceBenchmarkSettings) -> DataFrame:
    """Benchmark function with the discretized Hartmann 3D test function.

    Key characteristics:
    • Parameters:
      - x1: Discretisation of [0,1] with 25 steps
      - x2: Discretisation of [0,1] with 25 steps
      - x3: Discretisation of [0,1] with 25 steps
    • Output: Continuous
    • Objective: Minimization
    • Tests multiple recommenders:
      - Random Recommender
      - Default Recommender

    Args:
        settings: Configuration settings for the convergence benchmark.

    Returns:
        DataFrame containing benchmark results.
    """
    parameters = [
        NumericalDiscreteParameter("x1", np.linspace(0, 1, 25)),
        NumericalDiscreteParameter("x2", np.linspace(0, 1, 25)),
        NumericalDiscreteParameter("x3", np.linspace(0, 1, 25)),
    ]

    target = NumericalTarget(name="target", mode="MIN")
    searchspace = SearchSpace.from_product(parameters=parameters)
    objective = target.to_objective()

    scenarios: dict[str, Campaign] = {
        "Random Recommender": Campaign(
            searchspace=searchspace,
            recommender=RandomRecommender(),
            objective=objective,
        ),
        "Default Recommender": Campaign(
            searchspace=searchspace,
            objective=objective,
        ),
    }

    test_function = Hartmann(dim=3)

    lookup_discretized = arrays_to_dataframes(
        [p.name for p in parameters], [target.name], use_torch=True
    )(test_function)

    return simulate_scenarios(
        scenarios,
        lookup_discretized,
        batch_size=settings.batch_size,
        n_doe_iterations=settings.n_doe_iterations,
        n_mc_iterations=settings.n_mc_iterations,
        impute_mode="error",
    )


benchmark_config = ConvergenceBenchmarkSettings(
    batch_size=5,
    n_doe_iterations=30,
    n_mc_iterations=100,
)

hartmann_3d_discretized_benchmark = ConvergenceBenchmark(
    function=hartmann_3d_discretized,
    optimal_target_values={"target": -3.86278},
    settings=benchmark_config,
)



"""Validation functionality for constraints."""

from collections.abc import Collection
from itertools import combinations

from baybe.constraints.base import Constraint
from baybe.constraints.continuous import ContinuousCardinalityConstraint
from baybe.constraints.discrete import (
    DiscreteDependenciesConstraint,
)
from baybe.parameters import NumericalContinuousParameter
from baybe.parameters.base import Parameter

try:  # For python < 3.11, use the exceptiongroup backport
    ExceptionGroup
except NameError:
    from exceptiongroup import ExceptionGroup


def validate_constraints(  # noqa: DOC101, DOC103
    constraints: Collection[Constraint], parameters: Collection[Parameter]
) -> None:
    """Assert that a given collection of constraints is valid.

    Raises:
        ValueError: If there is more than one
            :class:`baybe.constraints.discrete.DiscreteDependenciesConstraint` declared.
        ValueError: If any two continuous cardinality constraints have an overlapping
            parameter set.
        ValueError: If any constraint contains an invalid parameter name.
        ValueError: If any continuous constraint includes a discrete parameter.
        ValueError: If any discrete constraint includes a continuous parameter.
        ValueError: If any discrete constraint that is valid only for numerical
            discrete parameters includes non-numerical discrete parameters.
        ValueError: If any parameter affected by a cardinality constraint does
            not include zero.
    """
    if sum(isinstance(itm, DiscreteDependenciesConstraint) for itm in constraints) > 1:
        raise ValueError(
            f"There is only one {DiscreteDependenciesConstraint.__name__} allowed. "
            f"Please specify all dependencies in one single constraint."
        )

    validate_cardinality_constraints_are_nonoverlapping(
        [con for con in constraints if isinstance(con, ContinuousCardinalityConstraint)]
    )

    param_names_all = [p.name for p in parameters]
    param_names_discrete = [p.name for p in parameters if p.is_discrete]
    param_names_continuous = [p.name for p in parameters if p.is_continuous]
    param_names_non_numerical = [p.name for p in parameters if not p.is_numerical]
    params_continuous: list[NumericalContinuousParameter] = [
        p for p in parameters if isinstance(p, NumericalContinuousParameter)
    ]

    for constraint in constraints:
        if not all(p in param_names_all for p in constraint.parameters):
            raise ValueError(
                f"You are trying to create a constraint with at least one parameter "
                f"name that does not exist in the list of defined parameters. "
                f"Parameter list of the affected constraint: {constraint.parameters}"
            )

        if constraint.is_continuous and any(
            p in param_names_discrete for p in constraint.parameters
        ):
            raise ValueError(
                f"You are trying to initialize a continuous constraint over a "
                f"parameter that is discrete. Parameter list of the affected "
                f"constraint: {constraint.parameters}"
            )

        if constraint.is_discrete and any(
            p in param_names_continuous for p in constraint.parameters
        ):
            raise ValueError(
                f"You are trying to initialize a discrete constraint over a parameter "
                f"that is continuous. Parameter list of the affected constraint: "
                f"{constraint.parameters}"
            )

        if constraint.numerical_only and any(
            p in param_names_non_numerical for p in constraint.parameters
        ):
            raise ValueError(
                f"You are trying to initialize a constraint of type "
                f"'{constraint.__class__.__name__}', which is valid only for numerical "
                f"discrete parameters, over a non-numerical parameter. "
                f"Parameter list of the affected constraint: {constraint.parameters}."
            )

        if isinstance(constraint, ContinuousCardinalityConstraint):
            validate_cardinality_constraint_parameter_bounds(
                constraint, params_continuous
            )


def validate_cardinality_constraints_are_nonoverlapping(
    constraints: Collection[ContinuousCardinalityConstraint],
) -> None:
    """Validate that cardinality constraints are non-overlapping.

    Args:
        constraints: A collection of continuous cardinality constraints.

    Raises:
        ValueError: If any two continuous cardinality constraints have an overlapping
            parameter set.
    """
    for c1, c2 in combinations(constraints, 2):
        if (s1 := set(c1.parameters)).intersection(s2 := set(c2.parameters)):
            raise ValueError(
                f"Constraints of type `{ContinuousCardinalityConstraint.__name__}` "
                f"cannot share the same parameters. Found the following overlapping "
                f"parameter sets: {s1}, {s2}."
            )


def validate_cardinality_constraint_parameter_bounds(
    constraint: ContinuousCardinalityConstraint,
    parameters: Collection[NumericalContinuousParameter],
) -> None:
    """Validate that all parameters of a continuous cardinality constraint include zero.

    Args:
        constraint: A continuous cardinality constraint.
        parameters: A collection of parameters, including those affected by the
            constraint.

    Raises:
        ValueError: If one of the affected parameters does not include zero.
        ExceptionGroup: If several of the affected parameters do not include zero.
    """
    exceptions = []
    for name in constraint.parameters:
        try:
            parameter = next(p for p in parameters if p.name == name)
        except StopIteration as ex:
            raise ValueError(
                f"The parameter '{name}' referenced by the constraint is not contained "
                f"in the given collection of parameters."
            ) from ex

        if not parameter.is_in_range(0.0):
            exceptions.append(
                ValueError(
                    f"The bounds of all parameters affected by a constraint of type "
                    f"'{ContinuousCardinalityConstraint.__name__}' must include zero, "
                    f"but the bounds of parameter '{name}' are "
                    f"{parameter.bounds.to_tuple()}, which may indicate unintended "
                    f"settings in your parameter definition. "
                    f"A parameter whose value range excludes zero trivially "
                    f"increases the cardinality of the resulting configuration by one. "
                    f"Therefore, if your parameter definitions are all correct, "
                    f"consider excluding the parameter from the constraint and "
                    f"reducing the cardinality limits by one accordingly."
                )
            )

    if exceptions:
        if len(exceptions) == 1:
            raise exceptions[0]
        raise ExceptionGroup("Invalid parameter bounds", exceptions)



## Single-Target vs. Pareto Optimization

# In this example, we illustrate the difference between single-target and Pareto
# optimization under laboratory conditions provided by a pair of synthetic targets.
# In particular:
# * We set up two quadratic target functions with different center points,
# * visualize the corresponding Pareto frontier,
# * and compare the recommendations obtained by optimizing the targets individually
#   and jointly.


### Imports

import os

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from baybe.campaign import Campaign
from baybe.objectives import ParetoObjective
from baybe.parameters import NumericalContinuousParameter
from baybe.searchspace import SearchSpace
from baybe.targets import NumericalTarget
from baybe.utils.dataframe import arrays_to_dataframes
from baybe.utils.random import set_random_seed

### Settings

# Let's first define some general settings for our example:

SMOKE_TEST = "SMOKE_TEST" in os.environ

BATCH_SIZE = 2 if SMOKE_TEST else 10
N_TRAINING_DATA = 2 if SMOKE_TEST else 100
N_GRID_POINTS = 3 if SMOKE_TEST else 100
CENTER_Y0 = (-0.5, -0.5)
CENTER_Y1 = (0.5, 0.5)

# Also, we fix the random seed for reproducibility:

set_random_seed(42)


### Defining the Optimization Problem

# We start by defining the parameters and targets describing the inputs and outputs
# of our synthetic functions:

x0 = NumericalContinuousParameter("x0", (-1, 1))
x1 = NumericalContinuousParameter("x1", (-1, 1))
y0 = NumericalTarget("y0", "MAX")
y1 = NumericalTarget("y1", "MIN")
searchspace = SearchSpace.from_product([x0, x1])

# With these definitions at hand, we can construct a multi-variate callable representing
# the two quadratic target functions:


@arrays_to_dataframes([x0.name, x1.name], [y0.name, y1.name])
def lookup(arr: np.ndarray) -> np.ndarray:
    """Compute (negative scaled) mean square values for different center points."""
    y0 = -np.sum((arr - CENTER_Y0) ** 2, axis=1)  # negative
    y1 = np.sum((arr - CENTER_Y1) ** 2, axis=1) * 100 + 1000  # scaled
    return np.c_[y0, y1]


### Campaign Setup

# We now query the callable with some randomly generated inputs to collect training
# data for our model:

data = searchspace.continuous.sample_uniform(N_TRAINING_DATA)
data = pd.concat([data, lookup(data)], axis=1)

# Next, we create three campaigns for comparison:
# * One focusing on the first target
# * One focusing on the second target
# * One for Pareto optimization of both targets

campaign_y0 = Campaign(searchspace=searchspace, objective=y0.to_objective())
campaign_y1 = Campaign(searchspace=searchspace, objective=y1.to_objective())
campaign_par = Campaign(searchspace=searchspace, objective=ParetoObjective([y0, y1]))

# We feed each campaign with the same training data and request recommendations:

campaign_y0.add_measurements(data)
campaign_y1.add_measurements(data)
campaign_par.add_measurements(data)

rec_y0 = campaign_y0.recommend(BATCH_SIZE)
rec_y1 = campaign_y1.recommend(BATCH_SIZE)
rec_par = campaign_par.recommend(BATCH_SIZE)

out_y0 = lookup(rec_y0)
out_y1 = lookup(rec_y1)
out_par = lookup(rec_par)


### Visualization

# To visualize the results, we first create grids to sample our target functions:

x0_mesh, x1_mesh = np.meshgrid(
    np.linspace(*x0.bounds.to_tuple(), N_GRID_POINTS),
    np.linspace(*x1.bounds.to_tuple(), N_GRID_POINTS),
)
df_y = lookup(pd.DataFrame({x0.name: x0_mesh.ravel(), x1.name: x1_mesh.ravel()}))
y0_mesh = np.reshape(df_y[y0.name], x0_mesh.shape)
y1_mesh = np.reshape(df_y[y1.name], x1_mesh.shape)


# Now, we can plot the function values, the training data, the recommendations,
# and the Pareto frontier in the parameter space:

fig, axs = plt.subplots(1, 2, figsize=(10, 5))

plt.sca(axs[0])
plt.contour(x0_mesh, x1_mesh, y0_mesh, colors="tab:red", alpha=0.2)
plt.contour(x0_mesh, x1_mesh, y1_mesh, colors="tab:blue", alpha=0.2)
plt.plot(*np.c_[CENTER_Y0, CENTER_Y1], "k", label="frontier")
plt.plot(data[x0.name], data[x1.name], "o", color="0.7", markersize=2, label="training")
plt.plot(rec_y0[x0.name], rec_y0[x1.name], "o", color="tab:red", label="single_y0")
plt.plot(rec_y1[x0.name], rec_y1[x1.name], "o", color="tab:blue", label="single_y1")
plt.plot(rec_par[x0.name], rec_par[x1.name], "o", color="tab:purple", label="pareto")
plt.legend(loc="upper left")
plt.xlabel(x0.name)
plt.ylabel(x1.name)
plt.title("Parameter Space")
plt.axis("square")
plt.axis([-1, 1, -1, 1])

# Similarly, we plot the training data, the achieved function values,
# and the Pareto frontier in the target space:

plt.sca(axs[1])
frontier = lookup(
    pd.DataFrame(np.linspace(CENTER_Y0, CENTER_Y1), columns=[x0.name, x1.name])
)
plt.plot(*frontier.to_numpy().T, "k", label="frontier")
plt.plot(data[y0.name], data[y1.name], "o", color="0.7", markersize=2, label="training")
plt.plot(out_y0[y0.name], out_y0[y1.name], "o", color="tab:red", label="single_y0")
plt.plot(out_y1[y0.name], out_y1[y1.name], "o", color="tab:blue", label="single_y1")
plt.plot(out_par[y0.name], out_par[y1.name], "o", color="tab:purple", label="pareto")
plt.legend(loc="upper left")
plt.xlabel(y0.name)
plt.ylabel(y1.name)
plt.title("Target Space")

plt.tight_layout()
if not SMOKE_TEST:
    plt.savefig("pareto.svg")



"""Categorical parameters."""

import gc
from functools import cached_property

import numpy as np
import pandas as pd
from attrs import Converter, define, field
from attrs.validators import deep_iterable, instance_of, min_len
from typing_extensions import override

from baybe.parameters.base import _DiscreteLabelLikeParameter
from baybe.parameters.enum import CategoricalEncoding
from baybe.parameters.validation import validate_unique_values
from baybe.utils.conversion import nonstring_to_tuple
from baybe.utils.numerical import DTypeFloatNumpy


def _convert_values(value, self, field) -> tuple[str, ...]:
    """Sort and convert values for categorical parameters."""
    value = nonstring_to_tuple(value, self, field)
    return tuple(sorted(value, key=lambda x: (str(type(x)), x)))


def _validate_label_min_len(self, attr, value) -> None:
    """An attrs-compatible validator to ensure minimum label length."""  # noqa: D401
    if isinstance(value, str) and len(value) < 1:
        raise ValueError(
            f"Strings used as '{attr.alias}' for '{self.__class__.__name__}' must "
            f"have at least 1 character."
        )


@define(frozen=True, slots=False)
class CategoricalParameter(_DiscreteLabelLikeParameter):
    """Parameter class for categorical parameters."""

    # object variables
    _values: tuple[str | bool, ...] = field(
        alias="values",
        converter=Converter(_convert_values, takes_self=True, takes_field=True),  # type: ignore
        validator=(  # type: ignore
            validate_unique_values,
            deep_iterable(
                member_validator=(instance_of((str, bool)), _validate_label_min_len),
                iterable_validator=min_len(2),
            ),
        ),
    )
    # See base class.

    encoding: CategoricalEncoding = field(
        default=CategoricalEncoding.OHE, converter=CategoricalEncoding
    )
    # See base class.

    @override
    @property
    def values(self) -> tuple:
        """The values of the parameter."""
        return self._values

    @override
    @cached_property
    def comp_df(self) -> pd.DataFrame:
        if self.encoding is CategoricalEncoding.OHE:
            cols = [
                f"{self.name}_{'b' if isinstance(val, bool) else ''}{val}"
                for val in self.values
            ]
            comp_df = pd.DataFrame(
                np.eye(len(self.values), dtype=DTypeFloatNumpy), columns=cols
            )
        elif self.encoding is CategoricalEncoding.INT:
            comp_df = pd.DataFrame(
                range(len(self.values)), dtype=DTypeFloatNumpy, columns=[self.name]
            )
        comp_df.index = pd.Index(self.values)

        return comp_df


@define(frozen=True, slots=False)
class TaskParameter(CategoricalParameter):
    """Parameter class for task parameters."""

    encoding: CategoricalEncoding = field(default=CategoricalEncoding.INT, init=False)
    # See base class.


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Tests for dataframe utilities."""

from contextlib import nullcontext

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal
from pytest import param

from baybe.exceptions import InputDataTypeWarning, SearchSpaceMatchWarning
from baybe.targets import BinaryTarget, NumericalTarget
from baybe.utils.dataframe import (
    add_noise_to_perturb_degenerate_rows,
    add_parameter_noise,
    fuzzy_row_match,
    handle_missing_values,
    normalize_input_dtypes,
)


@pytest.fixture()
def n_grid_points():
    return 5


def test_degenerate_rows():
    """Test noise-based deduplication of degenerate rows."""
    # Create random dataframe
    df = pd.DataFrame(np.random.randint(0, 100, size=(5, 3))).astype(float)

    # Manually create some degenerate rows
    df.loc[1] = df.loc[0]  # Make row 1 identical to row 0
    df.loc[3] = df.loc[2]  # Make row 3 identical to row 2
    df.iloc[:, -1] = 50.0  # Make last column constant to test the edge case

    # Add noise
    add_noise_to_perturb_degenerate_rows(df)

    # Assert that the utility fixed the degenerate rows
    assert not df.duplicated().any(), "Degenerate rows were not fixed by the utility."


def test_degenerate_rows_invalid_input():
    """Test that the utility correctly handles invalid input."""
    # Create random dataframe
    df = pd.DataFrame(np.random.randint(0, 100, size=(5, 3))).astype(float)

    # Manually create some degenerate rows
    df.loc[1] = df.loc[0]  # Make row 1 identical to row 0
    df.loc[3] = df.loc[2]  # Make row 3 identical to row 2

    # Insert invalid data types
    df = df.astype(object)  # to avoid pandas column dtype warnings
    df["invalid"] = "A"
    df.iloc[1, 0] = "B"

    # Add noise
    with pytest.raises(TypeError):
        add_noise_to_perturb_degenerate_rows(df)


@pytest.mark.parametrize(
    ("parameter_names", "noise", "duplicated"),
    [
        param(
            ["Categorical_1", "Num_disc_1", "Some_Setting"],
            False,
            True,
            id="discrete_num_noiseless_duplicated",
        ),
        param(
            ["Categorical_1", "Num_disc_1", "Some_Setting"],
            False,
            False,
            id="discrete_num_noiseless_unique",
        ),
        param(
            ["Categorical_1", "Num_disc_1", "Some_Setting"],
            True,
            False,
            id="discrete_num_noisy_unique",
        ),
        param(
            ["Categorical_1", "Switch_1", "Some_Setting"],
            False,
            False,
            id="discrete_cat",
        ),
        param(
            ["Categorical_1", "Switch_1", "Conti_finite1"],
            False,
            False,
            id="hybrid_cat",
        ),
        param(
            ["Categorical_1", "Num_disc_1", "Conti_finite1"],
            False,
            False,
            id="hybrid_num_noiseless_unique",
        ),
        param(
            ["Categorical_1", "Num_disc_1", "Conti_finite1"],
            True,
            False,
            id="hybrid_num_noisy_unique",
        ),
        param(
            ["Categorical_1", "Num_disc_1", "Conti_finite1"],
            False,
            True,
            id="hybrid_num_noiseless_duplicated",
        ),
    ],
)
def test_fuzzy_row_match(searchspace, noise, duplicated):
    """Fuzzy row matching returns expected indices."""
    left_df = searchspace.discrete.exp_rep.copy()
    selected = np.random.choice(left_df.index, 4, replace=False)
    right_df = left_df.loc[selected].reset_index(drop=True)

    context = nullcontext()
    if duplicated:
        # Set one of the input values to exactly the midpoint between two values to
        # cause a degenerate match
        vals = searchspace.get_parameters_by_name(["Num_disc_1"])[0].values
        right_df.loc[0, "Num_disc_1"] = vals[0] + (vals[1] - vals[0]) / 2.0
        context = pytest.warns(SearchSpaceMatchWarning, match="multiple matches")

    if noise:
        add_parameter_noise(
            right_df,
            searchspace.discrete.parameters,
            noise_type="relative_percent",
            noise_level=0.1,
        )

    with context as c:
        matched = fuzzy_row_match(left_df, right_df, searchspace.parameters)

    if duplicated:
        # Assert correct identification of problematic df parts
        w = next(x for x in c if isinstance(x.message, SearchSpaceMatchWarning)).message
        assert_frame_equal(right_df.loc[[0]], w.data)

        # Ignore problematic indices for subsequent equality check
        selected = selected[1:]
        matched = matched[1:]

    assert set(selected) == set(matched), (selected, matched)


@pytest.mark.parametrize(
    "parameter_names",
    [
        param(["Categorical_1", "Categorical_2", "Switch_1"], id="discrete"),
        param(["Categorical_1", "Num_disc_1", "Conti_finite1"], id="hybrid"),
    ],
)
@pytest.mark.parametrize("invalid", ["left", "right"])
def test_invalid_fuzzy_row_match(searchspace, invalid):
    """Returns expected errors when dataframes don't contain all expected columns."""
    left_df = searchspace.discrete.exp_rep.copy()
    selected = np.random.choice(left_df.index, 4, replace=False)
    right_df = left_df.loc[selected].copy()

    # Drop first column
    if invalid == "left":
        left_df = left_df.iloc[:, 1:]
    else:
        right_df = right_df.iloc[:, 1:]

    match = f"corresponding column in the {invalid} dataframe."
    with pytest.raises(ValueError, match=match):
        fuzzy_row_match(left_df, right_df, searchspace.parameters)


def test_measurement_singletons():
    """Correct treatment of rows with unmeasured targets."""
    df = pd.DataFrame(
        {
            "A": [np.nan, 2, 3, 4, 5],
            "B": ["a", "b", "b", "a", np.nan],
            "C": [55, 44, 33, 22, 11],
        },
        index=[123, 2, 3, 4, 5],
    )
    targets = [
        NumericalTarget(name="A", mode="MAX"),
        BinaryTarget(name="B", success_value="a", failure_value="b"),
    ]

    # Test NaN block
    with pytest.raises(
        ValueError, match=r"Bad input in the rows with these indices: \[123, 5\]"
    ):
        handle_missing_values(df, [t.name for t in targets])

    # Test NaN removal
    df_new = handle_missing_values(df, [t.name for t in targets], drop=True)
    assert_frame_equal(df.iloc[1:-1, :], df_new)


@pytest.mark.parametrize(
    "data, warning, match",
    [
        param(
            pd.DataFrame(
                {
                    "Num_disc_1": [1.1, 2.2],
                    "Target_max": [3.3, 4.4],
                }
            ),
            None,
            "",
            id="valid",
        ),
        param(
            pd.DataFrame(
                {
                    "Num_disc_1": [1.1, True],
                    "Target_max": [3.3, 4.4],
                }
            ),
            InputDataTypeWarning,
            "Num_disc_1",
            id="bool_in_num_parameter",
        ),
        param(
            pd.DataFrame(
                {
                    "Num_disc_1": [1.1, 2.2],
                    "Target_max": [3.3, False],
                }
            ),
            InputDataTypeWarning,
            "Target_max",
            id="bool_in_num_target",
        ),
    ],
)
def test_input_dtype(data, warning, match, parameters, targets):
    """If necessary, utility converts input dtype and raises warning."""
    with nullcontext() if warning is None else pytest.warns(warning, match=match):
        converted = normalize_input_dtypes(data, parameters + targets)
        converted = normalize_input_dtypes(data, parameters + targets)
        print(data)
        print(converted)

    # Asserts converted columns have expected dtypes
    assert pd.api.types.is_float_dtype(converted["Num_disc_1"]), (data, converted)
    assert pd.api.types.is_float_dtype(converted["Target_max"]), (data, converted)



"""Transfer learning benchmark with inverted Hartmann functions as tasks."""

from __future__ import annotations

import numpy as np
import pandas as pd
import torch
from botorch.test_functions.synthetic import Hartmann

from baybe.campaign import Campaign
from baybe.objectives import SingleTargetObjective
from baybe.parameters import NumericalDiscreteParameter, TaskParameter
from baybe.parameters.base import DiscreteParameter
from baybe.searchspace import SearchSpace
from baybe.simulation import simulate_scenarios
from baybe.targets import NumericalTarget
from baybe.utils.dataframe import arrays_to_dataframes
from benchmarks.definition import (
    ConvergenceBenchmark,
    ConvergenceBenchmarkSettings,
)


def hartmann_tl_3_20_15(settings: ConvergenceBenchmarkSettings) -> pd.DataFrame:
    """Benchmark function for transfer learning with the Hartmann function in 3D.

    Key characteristics:
    • Compares two versions of Hartmann function:
      - Target: standard Hartmann
      - Source: Hartmann with added noise (noise_std=0.15)
    • Uses 20 points per dimension
    • Tests transfer learning with different source data percentages:
      - 1% of source data
      - 10% of source data
      - 20% of source data

    Args:
        settings: Configuration settings for the convergence benchmark.

    Returns:
        DataFrame containing benchmark results.
    """
    target_function = Hartmann(dim=3)
    source_function = Hartmann(dim=3, noise_std=0.15)

    points_per_dim = 20
    percentages = [0.01, 0.05, 0.1]

    # Create grid locations for the parameters
    bounds = np.array([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]])
    grid_locations = {
        f"x{d}": np.linspace(lower, upper, points_per_dim)
        for d, (lower, upper) in enumerate(bounds.T)
    }

    params: list[DiscreteParameter] = [
        NumericalDiscreteParameter(
            name=name,
            values=points,
        )
        for name, points in grid_locations.items()
    ]
    task_param = TaskParameter(
        name="Function",
        values=["Target_Function", "Source_Function"],
        active_values=["Target_Function"],
    )
    params_tl = params + [task_param]

    searchspace_nontl = SearchSpace.from_product(parameters=params)
    searchspace_tl = SearchSpace.from_product(parameters=params_tl)

    objective = SingleTargetObjective(target=NumericalTarget(name="Target", mode="MIN"))
    tl_campaign = Campaign(
        searchspace=searchspace_tl,
        objective=objective,
    )
    nontl_campaign = Campaign(
        searchspace=searchspace_nontl,
        objective=objective,
    )

    meshgrid = np.meshgrid(*[points for points in grid_locations.values()])

    # Create a DataFrame for the initial data coordinates
    coord_columns = [p.name for p in params]
    initial_data = pd.DataFrame(
        {f"x{d}": grid_d.ravel() for d, grid_d in enumerate(meshgrid)},
        columns=coord_columns,  # Ensure correct column order
    )

    # Convert coordinates to a PyTorch tensor
    initial_data_tensor = torch.tensor(initial_data[coord_columns].values)

    target_values_tensor = source_function(initial_data_tensor)

    # Assign the results back to a new DataFrame for initial_data
    initial_data["Target"] = target_values_tensor.detach().numpy()
    initial_data["Function"] = "Source_Function"

    lookup = arrays_to_dataframes([p.name for p in params], ["Target"], use_torch=True)(
        target_function
    )

    results = []
    for p in percentages:
        results.append(
            simulate_scenarios(
                {
                    f"{int(100 * p)}": tl_campaign,
                    f"{int(100 * p)}_naive": nontl_campaign,
                },
                lookup,
                initial_data=[
                    initial_data.sample(frac=p) for _ in range(settings.n_mc_iterations)
                ],
                batch_size=settings.batch_size,
                n_doe_iterations=settings.n_doe_iterations,
                impute_mode="error",
            )
        )
    results.append(
        simulate_scenarios(
            {"0": tl_campaign, "0_naive": nontl_campaign},
            lookup,
            batch_size=settings.batch_size,
            n_doe_iterations=settings.n_doe_iterations,
            n_mc_iterations=settings.n_mc_iterations,
            impute_mode="error",
        )
    )
    return pd.concat(results)


benchmark_config = ConvergenceBenchmarkSettings(
    batch_size=2,
    n_doe_iterations=25,
    n_mc_iterations=75,
)

hartmann_tl_3_20_15_benchmark = ConvergenceBenchmark(
    function=hartmann_tl_3_20_15,
    optimal_target_values={"Target": -3.851831124860353},
    settings=benchmark_config,
)



"""Tests for partial measurements."""

from contextlib import nullcontext

import numpy as np
import pytest
from pytest import param

from baybe.acquisition import qLogNEHVI
from baybe.exceptions import IncompleteMeasurementsError
from baybe.objectives import DesirabilityObjective, ParetoObjective
from baybe.targets import NumericalTarget
from baybe.utils.dataframe import create_fake_input


@pytest.mark.parametrize(
    "objective",
    [
        param(NumericalTarget("t1", "MAX").to_objective(), id="single_target"),
        param(
            DesirabilityObjective(
                [
                    NumericalTarget("t1", "MAX", (0, 1)),
                    NumericalTarget("t2", "MIN", (0, 1)),
                ]
            ),
            id="desirability",
        ),
        param(
            ParetoObjective(
                [NumericalTarget("t1", "MAX"), NumericalTarget("t2", "MIN")],
            ),
            id="pareto",
        ),
    ],
)
@pytest.mark.parametrize("n_iterations", [2], ids=["i2"])
@pytest.mark.parametrize("batch_size", [5], ids=["b5"])
@pytest.mark.parametrize("n_grid_points", [5], ids=["g5"])
def test_partial_measurements(ongoing_campaign, objective):
    c = ongoing_campaign
    m = c.measurements
    ts = c.targets

    # Mark some measurements as unmeasured / failed
    m.iloc[0, m.columns.get_loc(ts[0].name)] = np.nan
    m.iloc[-1, m.columns.get_loc(ts[-1].name)] = np.nan
    c.update_measurements(m)

    # Failure is expected when single surrogates are used, which is exactly the case
    # for non-Pareto objectives. For ParetoObjectives, the filtering is expected to
    # remove affected rows, resulting in no error.
    context = nullcontext()
    if not isinstance(objective, ParetoObjective):
        context = pytest.raises(ValueError, match="Bad input in the rows")

    # Trigger refit
    with context:
        c.recommend(2)


@pytest.mark.parametrize(
    "acqf",
    [
        param(qLogNEHVI(), id="ref_point_default"),
        param(qLogNEHVI(reference_point=[0.4, 0.2]), id="ref_point_provided"),
    ],
)
@pytest.mark.parametrize(
    "objective",
    [
        param(
            ParetoObjective(
                [NumericalTarget("t1", "MAX"), NumericalTarget("t2", "MIN")],
            ),
            id="pareto",
        ),
    ],
)
@pytest.mark.parametrize("n_grid_points", [5], ids=["g5"])
def test_pareto_ref_point_from_incomplete_measurements(campaign):
    measurements = create_fake_input(
        campaign.parameters, campaign.targets, 2
    ).reset_index(drop=True)
    measurements.loc[0, "t1"] = None
    measurements.loc[1, "t2"] = None
    campaign.add_measurements(measurements)

    context = nullcontext()
    if campaign.recommender.recommender.acquisition_function.reference_point is None:
        context = pytest.raises(IncompleteMeasurementsError, match="at least one")
    with context:
        campaign.recommend(5)



"""Aryl halide benchmark for transfer learning.

Source task: 1-chloro-4-(trifluoromethyl)benzene
Target task: 1-iodo-4-methoxybenzene
"""

from __future__ import annotations

import pandas as pd

from benchmarks.definition import (
    ConvergenceBenchmark,
    ConvergenceBenchmarkSettings,
)
from benchmarks.domains.transfer_learning.aryl_halides.base import (
    aryl_halide_tl_substance_benchmark,
)


def aryl_halide_CT_IM_tl(
    settings: ConvergenceBenchmarkSettings,
) -> pd.DataFrame:
    """Benchmark function for transfer learning with aryl halide reactions.

    Key characteristics:
    • Compares transfer learning vs. non-transfer learning approaches
    • Source task: 1-chloro-4-(trifluoromethyl)benzene
    • Target task: 1-iodo-4-methoxybenzene
    • Tests varying amounts of source data:
      - 1% of source data
      - 5% of source data
      - 10% of source data
      - 20% of source data
    • Optimal parameters:
      - Base: "MTBD"
      - Ligand: "AdBrettPhos"
      - Additive: "N,N-dibenzylisoxazol-3-amine"
    • Optimal yield: 68.25%

    Args:
        settings: Configuration settings for the convergence benchmark

    Returns:
        DataFrame containing benchmark results
    """
    return aryl_halide_tl_substance_benchmark(
        settings=settings,
        source_tasks=["1-chloro-4-(trifluoromethyl)benzene"],
        target_tasks=["1-iodo-4-methoxybenzene"],
        percentages=[0.01, 0.05, 0.1, 0.2],
    )


benchmark_config = ConvergenceBenchmarkSettings(
    batch_size=2,
    n_doe_iterations=25,
    n_mc_iterations=50,
)

aryl_halide_CT_IM_tl_benchmark = ConvergenceBenchmark(
    function=aryl_halide_CT_IM_tl,
    optimal_target_values={"yield": 68.24812709999999},
    settings=benchmark_config,
)



"""Composite surrogate models."""

from __future__ import annotations

from collections.abc import Sequence
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Generic, Protocol, TypeVar

import pandas as pd
from attrs import define, field
from typing_extensions import override

from baybe.exceptions import IncompatibleSurrogateError
from baybe.objectives.base import Objective
from baybe.searchspace.core import SearchSpace
from baybe.serialization import converter
from baybe.serialization.mixin import SerialMixin
from baybe.surrogates.base import PosteriorStatistic, SurrogateProtocol
from baybe.surrogates.gaussian_process.core import GaussianProcessSurrogate
from baybe.utils.basic import is_all_instance
from baybe.utils.dataframe import handle_missing_values

if TYPE_CHECKING:
    from botorch.models.model import ModelList
    from botorch.posteriors import PosteriorList
    from torch import Tensor

_T = TypeVar("_T")


class _SurrogateGetter(Protocol):
    """A index-based mapping from strings to surrogates."""

    def __getitem__(self, key: str) -> SurrogateProtocol: ...


@define
class _ReplicationMapping(Generic[_T]):
    """A wrapper class providing copies of a given template object via indexing access.

    Essentially a serializable version of ``defaultdict(lambda: deepcopy(template))``.
    """

    template: _T = field()
    """The template object to be copied upon indexing access."""

    _data: dict[Any, _T] = field(init=False, factory=dict, eq=False)
    """An internal storage keeping track of already requested copies."""

    def __getitem__(self, key: Any, /) -> _T:
        """Create a new object copy upon first access."""
        if key not in self._data:
            self._data[key] = deepcopy(self.template)
        return self._data[key]


@define
class CompositeSurrogate(SerialMixin, SurrogateProtocol):
    """A class for composing multi-target surrogates from single-target surrogates."""

    # IMPROVE: Currently, the class is implemented in the most vanilla way, using only
    #   BayBE's existing interfaces. There are several ways how it can be
    #   further optimized by integrating it more directly with the underlying gpytorch
    #   models. However, this probably requires some additional code adaptations to
    #   achieve a full integration. Some future directions:
    #   * Instead of fitting the models sequentially, a parallel optimization can
    #     be done via `SumMarginalLogLikelihood`. However, a full integration would
    #     also require supporting different fitting routines (e.g. LOO)
    #   * The manual construction of the `PosteriorList` can be avoided when
    #     the posterior computation is triggered directly on the `ModelList`. However,
    #     this requires a clean integration of the necessary pre-processing steps
    #     (transformation to computational representation + scaling)
    #   * There is currently a lot of redundancy because each of the surrogates
    #     internally stores a references to the fitting context (e.g. search space,
    #     objective, ...)

    surrogates: _SurrogateGetter = field()
    """An index-based mapping from target names to single-target surrogates."""

    _target_names: tuple[str, ...] | None = field(init=False, eq=False)
    """The names of the targets modeled by the surrogate outputs."""

    @classmethod
    def from_replication(cls, surrogate: SurrogateProtocol) -> CompositeSurrogate:
        """Replicate a given single-target surrogate logic for multiple targets."""
        return CompositeSurrogate(_ReplicationMapping(surrogate))

    @property
    def _surrogates_flat(self) -> tuple[SurrogateProtocol, ...]:
        """The surrogates ordered according to the targets of the modeled objective."""
        assert self._target_names is not None
        return tuple(self.surrogates[t] for t in self._target_names)

    @override
    def fit(
        self,
        searchspace: SearchSpace,
        objective: Objective,
        measurements: pd.DataFrame,
    ) -> None:
        for target in objective.targets:
            # Drop partial measurements for the respective target
            measurements_filtered = handle_missing_values(
                measurements, [target.name], drop=True
            )

            self.surrogates[target.name].fit(
                searchspace, target.to_objective(), measurements_filtered
            )
        self._target_names = tuple(t.name for t in objective.targets)

    @override
    def to_botorch(self) -> ModelList:
        from botorch.models import ModelList
        from botorch.models.model_list_gp_regression import ModelListGP

        cls = (
            ModelListGP
            if is_all_instance(self._surrogates_flat, GaussianProcessSurrogate)
            else ModelList
        )
        return cls(*(s.to_botorch() for s in self._surrogates_flat))

    def posterior(self, candidates: pd.DataFrame) -> PosteriorList:
        """Compute the posterior for candidates in experimental representation.

        The (independent joint) posterior is represented as a collection of individual
        posterior models computed per target of the involved objective.
        For details, see :meth:`baybe.surrogates.base.Surrogate.posterior`.
        """
        if not all(hasattr(s, "posterior") for s in self._surrogates_flat):
            raise IncompatibleSurrogateError(
                f"'{self.__class__.__name__}' can only compute a posterior in "
                f"experimental representation if all involved surrogates offer "
                f"posteriors in experimental representation."
            )

        from botorch.posteriors import PosteriorList

        # TODO[typing]: a `has_all_attrs` typeguard similar to `is_all_instance` would
        #   be handy here but unclear if this is doable with the current typing system
        posteriors = [s.posterior(candidates) for s in self._surrogates_flat]  # type: ignore[attr-defined]
        return PosteriorList(*posteriors)

    def _posterior_comp(self, candidates_comp: Tensor, /) -> PosteriorList:
        """Compute the posterior for candidates in computational representation.

        The (independent joint) posterior is represented as a collection of individual
        posterior models computed per target of the involved objective.
        For details, see :meth:`baybe.surrogates.base.Surrogate._posterior_comp`.
        """
        if not all(hasattr(s, "_posterior_comp") for s in self._surrogates_flat):
            raise IncompatibleSurrogateError(
                f"'{self.__class__.__name__}' can only compute a posterior in "
                f"computational representation if all involved surrogates offer "
                f"posteriors in computational representation."
            )

        from botorch.posteriors import PosteriorList

        # TODO[typing]: a `has_all_attrs` typeguard similar to `is_all_instance` would
        #   be handy here but unclear if this is doable with the current typing system
        posteriors = [s._posterior_comp(candidates_comp) for s in self._surrogates_flat]  # type: ignore[attr-defined]
        return PosteriorList(*posteriors)

    def posterior_stats(
        self,
        candidates: pd.DataFrame,
        stats: Sequence[PosteriorStatistic] = ("mean", "std"),
    ) -> pd.DataFrame:
        """See :meth:`baybe.surrogates.base.Surrogate.posterior_stats`."""
        if not all(hasattr(s, "posterior_stats") for s in self._surrogates_flat):
            raise IncompatibleSurrogateError(
                "Posterior statistics can only be computed if all involved surrogates "
                "offer this computation."
            )

        dfs = [s.posterior_stats(candidates, stats) for s in self._surrogates_flat]  # type: ignore[attr-defined]
        return pd.concat(dfs, axis=1)


def _structure_surrogate_getter(obj: dict, _) -> _SurrogateGetter:
    """Resolve the object type."""
    if (type_ := obj.pop("type")) == _ReplicationMapping.__name__:
        return converter.structure(obj, _ReplicationMapping[SurrogateProtocol])
    elif type_ == "dict":
        return converter.structure(obj, dict[str, SurrogateProtocol])
    raise NotImplementedError(f"No structure hook implemented for '{type_}'.")


def _unstructure_surrogate_getter(obj: _SurrogateGetter) -> dict:
    """Add the object type information."""
    return {"type": type(obj).__name__, **converter.unstructure(obj)}


converter.register_structure_hook_func(
    lambda t: t is _SurrogateGetter, _structure_surrogate_getter
)
converter.register_unstructure_hook_func(
    lambda t: t is _SurrogateGetter, _unstructure_surrogate_getter
)



"""Optional ONNX import."""

from baybe.exceptions import OptionalImportError

try:
    import onnxruntime
except ModuleNotFoundError as ex:
    raise OptionalImportError(name="onnxruntime", group="onnx") from ex

__all__ = [
    "onnxruntime",
]



"""Executes the benchmarking module."""
# Run this via 'python -m benchmarks' from the root directory.

import argparse
import os
from collections.abc import Collection

from benchmarks.definition import Benchmark
from benchmarks.domains import BENCHMARKS
from benchmarks.persistence import (
    LocalFileObjectStorage,
    PathConstructor,
    S3ObjectStorage,
)
from benchmarks.result import Result

RUNS_IN_CI = "CI" in os.environ


def save_benchmark_data(benchmark: Benchmark, result: Result) -> None:
    """Save the benchmark data to the object storage."""
    path_constructor = PathConstructor.from_result(result)
    persist_dict = benchmark.to_dict() | result.to_dict()

    object_storage = S3ObjectStorage() if RUNS_IN_CI else LocalFileObjectStorage()
    object_storage.write_json(persist_dict, path_constructor)


def run_all_benchmarks() -> None:
    """Run all benchmarks."""
    for benchmark in BENCHMARKS:
        result = benchmark()
        save_benchmark_data(benchmark, result)


def run_benchmarks(benchmark_names: Collection[str]) -> None:
    """Run a subset based on the benchmark name."""
    for benchmark in BENCHMARKS:
        if benchmark.name not in benchmark_names:
            continue

        result = benchmark()
        save_benchmark_data(benchmark, result)


def main() -> None:
    """Execute the benchmarking module."""
    parser = argparse.ArgumentParser(description="Executes the benchmarking module.")
    parser.add_argument(
        "--benchmark-list", nargs="+", help="List of benchmarks to run", default=None
    )
    args = parser.parse_args()
    if not args.benchmark_list:
        run_all_benchmarks()
        return

    benchmark_execute_set = set(args.benchmark_list)
    run_benchmarks(benchmark_execute_set)


if __name__ == "__main__":
    main()



"""Optional chemistry imports."""

from baybe.exceptions import OptionalImportError

try:
    from rdkit import Chem
    from skfp import fingerprints
    from skfp.bases import BaseFingerprintTransformer
    from skfp.preprocessing import ConformerGenerator, MolFromSmilesTransformer

except ModuleNotFoundError as ex:
    raise OptionalImportError(name="scikit-fingerprints", group="chem") from ex

__all__ = [
    "Chem",
    "fingerprints",
    "BaseFingerprintTransformer",
    "ConformerGenerator",
    "MolFromSmilesTransformer",
]



"""Hartmann function in 6 dimensions as a benchmark."""

from __future__ import annotations

from botorch.test_functions.synthetic import Hartmann
from pandas import DataFrame

from baybe.campaign import Campaign
from baybe.parameters import NumericalContinuousParameter
from baybe.recommenders import RandomRecommender
from baybe.searchspace import SearchSpace
from baybe.simulation import simulate_scenarios
from baybe.targets import NumericalTarget
from baybe.utils.dataframe import arrays_to_dataframes
from benchmarks.definition.convergence import (
    ConvergenceBenchmark,
    ConvergenceBenchmarkSettings,
)


def hartmann_6d(settings: ConvergenceBenchmarkSettings) -> DataFrame:
    """Benchmark function with the Hartmann 6D test function.

    Key characteristics:
    • Parameters:
      - x1: Continuous [0, 1]
      - x2: Continuous [0, 1]
      - x3: Continuous [0, 1]
      - x4: Continuous [0, 1]
      - x5: Continuous [0, 1]
      - x6: Continuous [0, 1]
    • Output: Continuous
    • Objective: Minimization
    • Optimal inputs:
        {x1: 0.20169, x2: 0.150011, x3: 0.476874,
        x4: 0.275332, x5: 0.311652, x6: 0.6573}
    • Optimal output: -3.86278
    • Tests multiple recommenders:
      - Random Recommender
      - Default Recommender

    Args:
        settings: Configuration settings for the convergence benchmark

    Returns:
        DataFrame containing benchmark results
    """
    parameters = [
        NumericalContinuousParameter("x1", (0.0, 1.0)),
        NumericalContinuousParameter("x2", (0.0, 1.0)),
        NumericalContinuousParameter("x3", (0.0, 1.0)),
        NumericalContinuousParameter("x4", (0.0, 1.0)),
        NumericalContinuousParameter("x5", (0.0, 1.0)),
        NumericalContinuousParameter("x6", (0.0, 1.0)),
    ]

    target = NumericalTarget(name="target", mode="MIN")
    searchspace = SearchSpace.from_product(parameters=parameters)
    objective = target.to_objective()

    scenarios: dict[str, Campaign] = {
        "Random Recommender": Campaign(
            searchspace=searchspace,
            recommender=RandomRecommender(),
            objective=objective,
        ),
        "Default Recommender": Campaign(
            searchspace=searchspace,
            objective=objective,
        ),
    }

    hartmann = Hartmann(dim=6)

    lookup = arrays_to_dataframes(
        [p.name for p in parameters], [target.name], use_torch=True
    )(hartmann)

    return simulate_scenarios(
        scenarios,
        lookup,
        batch_size=settings.batch_size,
        n_doe_iterations=settings.n_doe_iterations,
        n_mc_iterations=settings.n_mc_iterations,
        impute_mode="error",
    )


benchmark_config = ConvergenceBenchmarkSettings(
    batch_size=5,
    n_doe_iterations=30,
    n_mc_iterations=100,
)

hartmann_6d_benchmark = ConvergenceBenchmark(
    function=hartmann_6d,
    optimal_target_values={"target": -3.86278},
    settings=benchmark_config,
)



"""Hypothesis strategies for acquisition functions."""

import hypothesis.strategies as st

from baybe.acquisition import (
    ExpectedImprovement,
    LogExpectedImprovement,
    PosteriorMean,
    PosteriorStandardDeviation,
    ProbabilityOfImprovement,
    UpperConfidenceBound,
    qExpectedImprovement,
    qKnowledgeGradient,
    qLogExpectedImprovement,
    qLogNoisyExpectedHypervolumeImprovement,
    qLogNoisyExpectedImprovement,
    qNegIntegratedPosteriorVariance,
    qNoisyExpectedHypervolumeImprovement,
    qNoisyExpectedImprovement,
    qPosteriorStandardDeviation,
    qProbabilityOfImprovement,
    qSimpleRegret,
    qUpperConfidenceBound,
)
from baybe.utils.sampling_algorithms import DiscreteSamplingMethod

from ..hypothesis_strategies.basic import finite_floats


@st.composite
def _qNIPV_strategy(draw: st.DrawFn):
    sampling_fraction = draw(
        st.one_of(
            finite_floats(min_value=0.0, max_value=1.0, exclude_min=True),
            st.none(),
        )
    )

    sampling_n_points = None
    if sampling_fraction is None:
        sampling_n_points = draw(st.one_of(st.none(), st.integers(min_value=1)))

    return qNegIntegratedPosteriorVariance(
        sampling_fraction=sampling_fraction,
        sampling_n_points=sampling_n_points,
        sampling_method=draw(st.sampled_from(DiscreteSamplingMethod)),
    )


@st.composite
def _reference_points(draw: st.DrawFn):
    """Draw reference points for hypervolume improvement acquisition functions."""
    if draw(st.booleans()):
        return draw(st.lists(finite_floats(), min_size=1))
    return draw(finite_floats())


# These acqfs are ordered roughly according to increasing complexity
acquisition_functions = st.one_of(
    st.builds(ExpectedImprovement),
    st.builds(ProbabilityOfImprovement),
    st.builds(UpperConfidenceBound, beta=finite_floats()),
    st.builds(PosteriorMean),
    st.builds(PosteriorStandardDeviation, maximize=st.sampled_from([True, False])),
    st.builds(qPosteriorStandardDeviation),
    st.builds(LogExpectedImprovement),
    st.builds(qExpectedImprovement),
    st.builds(qProbabilityOfImprovement),
    st.builds(qUpperConfidenceBound, beta=finite_floats()),
    st.builds(qSimpleRegret),
    st.builds(qLogExpectedImprovement),
    st.builds(
        qKnowledgeGradient, num_fantasies=st.integers(min_value=1, max_value=512)
    ),
    st.builds(qNoisyExpectedImprovement, prune_baseline=st.booleans()),
    st.builds(qLogNoisyExpectedImprovement, prune_baseline=st.booleans()),
    _qNIPV_strategy(),
    st.builds(
        qNoisyExpectedHypervolumeImprovement,
        prune_baseline=st.booleans(),
        reference_point=_reference_points(),
    ),
    st.builds(
        qLogNoisyExpectedHypervolumeImprovement,
        prune_baseline=st.booleans(),
        reference_point=_reference_points(),
    ),
)



"""Continuous constraints."""

from __future__ import annotations

import gc
import math
from collections.abc import Collection, Iterator, Sequence
from itertools import combinations
from math import comb
from typing import TYPE_CHECKING, Any

import numpy as np
from attr.validators import gt, in_, lt
from attrs import define, field

from baybe.constraints.base import (
    CardinalityConstraint,
    ContinuousConstraint,
    ContinuousNonlinearConstraint,
)
from baybe.parameters import NumericalContinuousParameter
from baybe.utils.interval import Interval
from baybe.utils.numerical import DTypeFloatNumpy
from baybe.utils.validation import finite_float

if TYPE_CHECKING:
    from torch import Tensor

_valid_linear_constraint_operators = ["=", ">=", "<="]


@define
class ContinuousLinearConstraint(ContinuousConstraint):
    """Class for continuous linear constraints.

    Continuous linear constraints use parameter lists and coefficients to define
    in-/equality constraints over a continuous parameter space.
    """

    # object variables
    operator: str = field(validator=in_(_valid_linear_constraint_operators))
    """Defines the operator used in the equation. Internally this will negate rhs and
    coefficients for `<=`."""

    coefficients: list[float] = field()
    """In-/equality coefficient for each entry in ``parameters``."""

    rhs: float = field(default=0.0, converter=float, validator=finite_float)
    """Right-hand side value of the in-/equality."""

    @coefficients.validator
    def _validate_coefficients(  # noqa: DOC101, DOC103
        self, _: Any, coefficients: list[float]
    ) -> None:
        """Validate the coefficients.

        Raises:
            ValueError: If the number of coefficients does not match the number of
                parameters.
        """
        if len(self.parameters) != len(coefficients):
            raise ValueError(
                "The given 'coefficients' list must have one floating point entry for "
                "each entry in 'parameters'."
            )

    @coefficients.default
    def _default_coefficients(self):
        """Return equal weight coefficients as default."""
        return [1.0] * len(self.parameters)

    @property
    def _multiplier(self) -> float:
        """The internal multiplier for rhs and coefficients."""
        return -1.0 if self.operator == "<=" else 1.0

    @property
    def is_eq(self):
        """Whether this constraint models an equality (assumed inequality otherwise)."""
        return self.operator == "="

    def _drop_parameters(
        self, parameter_names: Collection[str]
    ) -> ContinuousLinearConstraint:
        """Create a copy of the constraint with certain parameters removed.

        Args:
            parameter_names: The names of the parameter to be removed.

        Returns:
            The reduced constraint.
        """
        parameters = [p for p in self.parameters if p not in parameter_names]
        coefficients = [
            c
            for p, c in zip(self.parameters, self.coefficients)
            if p not in parameter_names
        ]
        return ContinuousLinearConstraint(
            parameters, self.operator, coefficients, self.rhs
        )

    def to_botorch(
        self, parameters: Sequence[NumericalContinuousParameter], idx_offset: int = 0
    ) -> tuple[Tensor, Tensor, float]:
        """Cast the constraint in a format required by botorch.

        Used in calling ``optimize_acqf_*`` functions, for details see
        :func:`botorch.optim.optimize.optimize_acqf`

        Args:
            parameters: The parameter objects of the continuous space.
            idx_offset: Offset to the provided parameter indices.

        Returns:
            The tuple required by botorch.
        """
        import torch

        from baybe.utils.torch import DTypeFloatTorch

        param_names = [p.name for p in parameters]
        param_indices = [
            param_names.index(p) + idx_offset
            for p in self.parameters
            if p in param_names
        ]

        return (
            torch.tensor(param_indices),
            torch.tensor(
                [self._multiplier * c for c in self.coefficients], dtype=DTypeFloatTorch
            ),
            np.asarray(self._multiplier * self.rhs, dtype=DTypeFloatNumpy).item(),
        )


@define
class ContinuousCardinalityConstraint(
    CardinalityConstraint, ContinuousNonlinearConstraint
):
    """Class for continuous cardinality constraints."""

    relative_threshold: float = field(
        default=1e-3, converter=float, validator=[gt(0.0), lt(1.0)]
    )
    """A relative threshold for determining if a value is considered zero.

    The threshold is translated into an asymmetric open interval around zero via
    :meth:`get_absolute_thresholds`.

    **Note:** The interval induced by the threshold is considered **open** because
    numerical routines that optimize parameter values on the complementary set (i.e. the
    value range considered "nonzero") may push the numerical value exactly to the
    interval boundary, which should therefore also be considered "nonzero".
    """

    @property
    def n_inactive_parameter_combinations(self) -> int:
        """The number of possible inactive parameter combinations."""
        return sum(
            comb(len(self.parameters), n_inactive_parameters)
            for n_inactive_parameters in self._inactive_set_sizes()
        )

    def _inactive_set_sizes(self) -> range:
        """Get all possible sizes of inactive parameter sets."""
        return range(
            len(self.parameters) - self.max_cardinality,
            len(self.parameters) - self.min_cardinality + 1,
        )

    def inactive_parameter_combinations(self) -> Iterator[frozenset[str]]:
        """Get an iterator over all possible combinations of inactive parameters."""
        for n_inactive_parameters in self._inactive_set_sizes():
            yield from combinations(self.parameters, n_inactive_parameters)

    def sample_inactive_parameters(self, batch_size: int = 1) -> list[set[str]]:
        """Sample sets of inactive parameters according to the cardinality constraints.

        Args:
            batch_size: The number of parameter sets to be sampled.

        Returns:
            A list of sampled inactive parameter sets, where each set holds the
            corresponding parameter names.
        """
        # The number of possible parameter configuration per set cardinality
        n_configurations_per_cardinality = [
            math.comb(len(self.parameters), n)
            for n in range(self.min_cardinality, self.max_cardinality + 1)
        ]

        # Probability of each set cardinality under the assumption that all possible
        # inactive parameter sets are equally likely
        probabilities = n_configurations_per_cardinality / np.sum(
            n_configurations_per_cardinality
        )

        # Sample the number of active/inactive parameters
        n_active_params = np.random.choice(
            np.arange(self.min_cardinality, self.max_cardinality + 1),
            batch_size,
            p=probabilities,
        )
        n_inactive_params = len(self.parameters) - n_active_params

        # Sample the inactive parameters
        inactive_params = [
            set(np.random.choice(self.parameters, n_inactive, replace=False))
            for n_inactive in n_inactive_params
        ]

        return inactive_params

    def get_absolute_thresholds(self, bounds: Interval, /) -> Interval:
        """Get the absolute thresholds for a given interval.

        Turns the relative threshold of the constraint into absolute thresholds
        for the considered interval. That is, for a given interval ``(a, b)`` with
        ``a <= 0`` and ``b >= 0``, the method returns the interval ``(r*a, r*b)``,
        where ``r`` is the relative threshold defined by the constraint.

        Args:
            bounds: The specified interval.

        Returns:
            The absolute thresholds represented as an interval.

        Raises:
            ValueError: When the specified interval does not contain zero.
        """
        if not bounds.contains(0.0):
            raise ValueError(
                f"The specified interval must contain zero. Given: {bounds.to_tuple()}."
            )

        return Interval(
            lower=self.relative_threshold * bounds.lower,
            upper=self.relative_threshold * bounds.upper,
        )


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""EDBO preset for Gaussian process surrogates."""

from __future__ import annotations

import gc
from collections.abc import Collection
from typing import TYPE_CHECKING

from attrs import define
from typing_extensions import override

from baybe.kernels.basic import MaternKernel
from baybe.kernels.composite import ScaleKernel
from baybe.parameters import TaskParameter
from baybe.parameters.enum import SubstanceEncoding
from baybe.parameters.substance import SubstanceParameter
from baybe.priors.basic import GammaPrior
from baybe.searchspace.discrete import SubspaceDiscrete
from baybe.surrogates.gaussian_process.kernel_factory import KernelFactory

if TYPE_CHECKING:
    from torch import Tensor

    from baybe.kernels.base import Kernel
    from baybe.searchspace.core import SearchSpace


def _contains_encoding(
    subspace: SubspaceDiscrete, encodings: Collection[SubstanceEncoding]
) -> bool:
    """Tell if any of the substance parameters uses one of the specified encodings."""
    return any(
        p.encoding in encodings
        for p in subspace.parameters
        if isinstance(p, SubstanceParameter)
    )


_EDBO_ENCODINGS = (
    SubstanceEncoding.MORDRED,
    SubstanceEncoding.RDKIT,
    SubstanceEncoding.RDKIT2DDESCRIPTORS,
)
"""Encodings relevant to EDBO logic."""


@define
class EDBOKernelFactory(KernelFactory):
    """A factory providing the kernel for Gaussian process surrogates adapted from EDBO.

    References:
        * https://github.com/b-shields/edbo/blob/master/edbo/bro.py#L664
        * https://doi.org/10.1038/s41586-021-03213-y
    """

    @override
    def __call__(
        self, searchspace: SearchSpace, train_x: Tensor, train_y: Tensor
    ) -> Kernel:
        effective_dims = train_x.shape[-1] - len(
            [p for p in searchspace.parameters if isinstance(p, TaskParameter)]
        )

        switching_condition = _contains_encoding(
            searchspace.discrete, _EDBO_ENCODINGS
        ) and (effective_dims >= 50)

        # low D priors
        if effective_dims < 5:
            lengthscale_prior = GammaPrior(1.2, 1.1)
            lengthscale_initial_value = 0.2
            outputscale_prior = GammaPrior(5.0, 0.5)
            outputscale_initial_value = 8.0

        # DFT optimized priors
        elif switching_condition and effective_dims < 100:
            lengthscale_prior = GammaPrior(2.0, 0.2)
            lengthscale_initial_value = 5.0
            outputscale_prior = GammaPrior(5.0, 0.5)
            outputscale_initial_value = 8.0

        # Mordred optimized priors
        elif switching_condition:
            lengthscale_prior = GammaPrior(2.0, 0.1)
            lengthscale_initial_value = 10.0
            outputscale_prior = GammaPrior(2.0, 0.1)
            outputscale_initial_value = 10.0

        # OHE optimized priors
        else:
            lengthscale_prior = GammaPrior(3.0, 1.0)
            lengthscale_initial_value = 2.0
            outputscale_prior = GammaPrior(5.0, 0.2)
            outputscale_initial_value = 20.0

        return ScaleKernel(
            MaternKernel(
                nu=2.5,
                lengthscale_prior=lengthscale_prior,
                lengthscale_initial_value=lengthscale_initial_value,
            ),
            outputscale_prior=outputscale_prior,
            outputscale_initial_value=outputscale_initial_value,
        )


def _edbo_noise_factory(
    searchspace: SearchSpace, train_x: Tensor, train_y: Tensor
) -> tuple[GammaPrior, float]:
    """Create the default noise settings for the Gaussian process surrogate.

    The logic is adapted from EDBO (Experimental Design via Bayesian Optimization).

    References:
        * https://github.com/b-shields/edbo/blob/master/edbo/bro.py#L664
        * https://doi.org/10.1038/s41586-021-03213-y
    """
    # TODO: Replace this function with a proper likelihood factory
    effective_dims = train_x.shape[-1] - len(
        [p for p in searchspace.parameters if isinstance(p, TaskParameter)]
    )

    switching_condition = _contains_encoding(
        searchspace.discrete, _EDBO_ENCODINGS
    ) and (effective_dims >= 50)

    # low D priors
    if effective_dims < 5:
        return (GammaPrior(1.05, 0.5), 0.1)

    # DFT optimized priors
    elif switching_condition and effective_dims < 100:
        return (GammaPrior(1.5, 0.1), 5.0)

    # Mordred optimized priors
    elif switching_condition:
        return (GammaPrior(1.5, 0.1), 5.0)

    # OHE optimized priors
    else:
        return (GammaPrior(1.5, 0.1), 5.0)


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""BayBE — A Bayesian Back End for Design of Experiments."""

import warnings

from baybe.campaign import Campaign

# Show deprecation warnings
warnings.filterwarnings("default", category=DeprecationWarning, module="baybe")


def infer_version() -> str:  # pragma: no cover
    """Determine the package version for the different ways the code can be invoked."""
    # ----------------------------------------------------------------------------------
    # Attempt 1:
    # Try to dynamically infer the version on the fly from the git history using
    # setuptools_scm, provided the latter is installed. This relies on the same logic as
    # for writing the version metadata during the installation process (see Attempt 2
    # below) but has the advantage that the version info is always up-to-date, even when
    # using an "editable installation" using the "-e" flag, where the metadata file may
    # become outdated when the git HEAD changes.
    try:
        from pathlib import Path

        from setuptools_scm import get_version

        return get_version(
            root=str(Path(__path__[0]).parent),
            version_scheme="post-release",
            local_scheme="dirty-tag",
        )
    except (LookupError, ImportError):
        pass

    # ----------------------------------------------------------------------------------
    # Attempt 2:
    # If the package is installed, retrieve the version from the metadata generated
    # via setuptools_scm during the installation process. (If the install had been
    # attempted without the git folder in place, setuptools_scm would have complained,
    # causing the installation process to fail.)
    from importlib.metadata import version

    try:
        return version(__name__)
    except ModuleNotFoundError:
        pass

    # ----------------------------------------------------------------------------------
    # Fallback:
    # If neither the package is installed nor the git history is available, return
    # "unknown" since the version cannot be determined.
    return "unknown"


__version__ = infer_version()
__all__ = [
    "__version__",
    "Campaign",
]

del infer_version



"""Custom exceptions and warnings."""

from typing import Any

import pandas as pd
from attr.validators import instance_of
from attrs import define, field
from typing_extensions import override

##### Warnings #####


class InputDataTypeWarning(UserWarning):
    """An input has unexpected data type."""


class UnusedObjectWarning(UserWarning):
    """
    A method or function was called with undesired arguments which indicates an
    unintended user fault.
    """


@define
class SearchSpaceMatchWarning(UserWarning):
    """
    When trying to match data to entries in the search space, something unexpected
    happened.
    """

    message: str = field(validator=instance_of(str))
    data: pd.DataFrame = field(validator=instance_of(pd.DataFrame))

    def __attrs_pre_init(self):
        super().__init__(self.message)

    @override
    def __str__(self):
        return self.message


class MinimumCardinalityViolatedWarning(UserWarning):
    """Minimum cardinality constraints are violated."""


##### Exceptions #####


class IncompatibilityError(Exception):
    """Incompatible components are used together."""


class IncompatibleSearchSpaceError(IncompatibilityError):
    """
    A recommender is used with a search space that contains incompatible parts,
    e.g. a discrete recommender is used with a hybrid or continuous search space.
    """


class IncompatibleSurrogateError(IncompatibilityError):
    """An incompatible surrogate was selected."""


class IncompatibleAcquisitionFunctionError(IncompatibilityError):
    """An incompatible acquisition function was selected."""


class IncompatibleExplainerError(IncompatibilityError):
    """An explainer is incompatible with the data it is presented."""


class IncompatibleArgumentError(IncompatibilityError):
    """An incompatible argument was passed to a callable."""


class InfeasibilityError(Exception):
    """An optimization problem has no feasible solution."""


class NotEnoughPointsLeftError(Exception):
    """
    More recommendations are requested than there are viable parameter configurations
    left in the search space.
    """


class NoMCAcquisitionFunctionError(Exception):
    """
    A Monte Carlo acquisition function is required but an analytical acquisition
    function has been selected by the user.
    """


class EmptySearchSpaceError(Exception):
    """The created search space contains no parameters."""


class NoMeasurementsError(Exception):
    """A context expected measurements but none were available."""


class IncompleteMeasurementsError(Exception):
    """A context expected complete measurements but none were available."""


class NothingToSimulateError(Exception):
    """There is nothing to simulate because there are no testable configurations."""


class NoRecommendersLeftError(Exception):
    """A recommender is requested by a meta recommender but there are no recommenders
    left.
    """


class NumericalUnderflowError(Exception):
    """A computation would lead to numerical underflow."""


class OptionalImportError(ImportError):
    """An attempt was made to import an optional but uninstalled dependency."""

    def __init__(
        self,
        *args: Any,
        name: str | None = None,
        path: str | None = None,
        group: str | None = None,
    ):
        super().__init__(*args, name=name, path=path)

        # If no message has been explicitly set, create it from the context
        if self.msg is None and name is not None:
            group_str = f"`pip install 'baybe[{group}]'` or " if group else ""
            self.msg = (
                f"The requested functionality requires the optional "
                f"'{self.name}' package, which is currently not installed. "
                f"Please install the dependency and try again. "
                f"You can do so manually (e.g. `pip install {self.name}`) "
                f"or using an appropriate optional dependency group "
                f"(e.g. {group_str}`pip install 'baybe[extras]'`)."
            )


class DeprecationError(Exception):
    """Signals the use of a deprecated mechanism to the user, interrupting execution."""


class UnidentifiedSubclassError(Exception):
    """A specified subclass cannot be found in the given class hierarchy."""


class ModelNotTrainedError(Exception):
    """A prediction/transformation is attempted before the model has been trained."""


class UnmatchedAttributeError(Exception):
    """An attribute cannot be matched against a certain callable signature."""


class InvalidSurrogateModelError(Exception):
    """An invalid surrogate model was chosen."""


class InvalidTargetValueError(Exception):
    """A target value was entered that is not in the target space."""



"""Tests for custom surrogate models."""

from contextlib import nullcontext

import pytest

from baybe import Campaign
from baybe._optional.info import ONNX_INSTALLED
from baybe.surrogates import CustomONNXSurrogate
from tests.conftest import run_iterations


@pytest.mark.skipif(
    not ONNX_INSTALLED, reason="Optional onnx dependency not installed."
)
def test_invalid_onnx_creation(onnx_str):
    """Invalid onnx model creation."""
    # Scenario: No input
    with pytest.raises(TypeError):
        CustomONNXSurrogate()

    # Scenario: No onnx input name
    with pytest.raises(TypeError):
        CustomONNXSurrogate(onnx_str=b"onnx_str")

    # Scenario: No onnx str
    with pytest.raises(TypeError):
        CustomONNXSurrogate(onnx_input_name="input")


@pytest.mark.skipif(
    not ONNX_INSTALLED, reason="Optional onnx dependency not installed."
)
@pytest.mark.parametrize("surrogate_model", ["onnx"], indirect=True)
@pytest.mark.parametrize(
    ["parameter_names", "should_raise"],
    [(["Categorical_1"], True), (["Some_Setting"], False)],
)
def test_supported_parameter_types(campaign: Campaign, should_raise: bool):
    """Using an ONNX model with unsupported parameters should raise an exception."""
    run_iterations(campaign, n_iterations=1, batch_size=1)
    context = pytest.raises(TypeError) if should_raise else nullcontext()
    with context:
        campaign.recommend(batch_size=1)



"""Target lookup mechanisms."""

from __future__ import annotations

import logging
from collections.abc import Callable, Collection
from typing import Literal

import numpy as np
import pandas as pd

from baybe.simulation._imputation import _impute_lookup
from baybe.targets.base import Target
from baybe.utils.dataframe import add_fake_measurements

_logger = logging.getLogger(__name__)


def look_up_targets(
    queries: pd.DataFrame,
    targets: Collection[Target],
    lookup: pd.DataFrame | Callable[[pd.DataFrame], pd.DataFrame] | None,
    impute_mode: Literal[
        "error", "worst", "best", "mean", "random", "ignore"
    ] = "error",
) -> None:
    """Add/fill target values in a dataframe using a lookup mechanism.

    Note:
        This does not create a new dataframe but modifies ``queries`` in-place.

    Args:
        queries: The dataframe to be modified. Its content must be compatible with the
            chosen lookup mechanism.
        targets: The targets whose values are to be looked up.
        lookup: The lookup mechanism. Can be one of the following choices:

            -   A dataframe mapping rows of ``queries`` to the corresponding target
                values. That is, it must contain the same columns as ``queries`` plus
                one additional column for each of the given target.
            -   A callable, providing target values for each row of ``queries``.
            -   ``None``. Produces fake values for all targets.
        impute_mode: Specifies how a missing lookup will be handled. Only relevant for
            dataframe lookups. Can be one of the following choices:

            - ``"error"``: An error will be thrown.
            - ``"worst"``: Imputes the worst available value for each target.
            - ``"best"``: Imputes the best available value for each target.
            - ``"mean"``: Imputes the mean value for each target.
            - ``"random"``: A random row will be used for the lookup.

    Raises:
        ValueError: If an unsupported lookup mechanism is provided.

    Example:
        >>> import pandas as pd
        >>> from baybe.targets.numerical import NumericalTarget
        >>> from baybe.simulation.lookup import look_up_targets
        >>>
        >>> targets = [NumericalTarget("target", "MAX")]
        >>> df = pd.DataFrame({"x": [1, 2, 3]})
        >>> lookup_df = pd.DataFrame({"x": [1, 2], "target": [10, 20]})
        >>> look_up_targets(df, targets, lookup_df, impute_mode="mean")
        >>> print(df)
           x  target
        0  1    10.0
        1  2    20.0
        2  3    15.0
    """
    if lookup is None:
        add_fake_measurements(queries, targets)
    elif isinstance(lookup, Callable):
        _look_up_targets_from_callable(queries, lookup)
    elif isinstance(lookup, pd.DataFrame):
        _look_up_targets_from_dataframe(queries, targets, lookup, impute_mode)
    else:
        raise ValueError("Unsupported lookup mechanism.")


def _look_up_targets_from_callable(
    queries: pd.DataFrame, lookup: Callable[[pd.DataFrame], pd.DataFrame]
) -> None:
    """Look up target values by querying a callable."""
    df_targets = lookup(queries)
    queries[df_targets.columns] = df_targets


def _look_up_targets_from_dataframe(
    queries: pd.DataFrame,
    targets: Collection[Target],
    lookup: pd.DataFrame,
    impute_mode: Literal[
        "error", "worst", "best", "mean", "random", "ignore"
    ] = "error",
) -> None:
    """Look up target values from a dataframe."""
    # # IMPROVE: Although it's not too important for a simulation, this
    # #  could also be implemented for approximate matches"""

    target_names = [t.name for t in targets]

    all_match_vals = []
    for _, row in queries.iterrows():
        # IMPROVE: to the entire matching at once via a merge
        ind = lookup[
            (lookup.loc[:, row.index] == row).all(axis=1, skipna=False)
        ].index.values

        if len(ind) > 1:
            # More than two instances of this parameter combination
            # have been measured
            _logger.warning(
                "The lookup rows with indexes %s seem to be "
                "duplicates regarding parameter values. Choosing a "
                "random one.",
                ind,
            )
            match_vals = lookup.loc[np.random.choice(ind), target_names].values

        elif len(ind) < 1:
            # Parameter combination cannot be looked up and needs to be
            # imputed.
            if impute_mode == "ignore":
                raise AssertionError(
                    "Something went wrong for impute_mode 'ignore'. "
                    "It seems the search space was not correctly "
                    "reduced before recommendations were generated."
                )
            match_vals = _impute_lookup(row, lookup, targets, impute_mode)

        else:
            # Exactly one match has been found
            match_vals = lookup.loc[ind[0], target_names].values

        # Collect the matches
        all_match_vals.append(match_vals)

    # Add the lookup values
    queries.loc[:, target_names] = np.asarray(all_match_vals)



"""Preset configurations for Gaussian process surrogates."""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from baybe.surrogates.gaussian_process.core import GaussianProcessSurrogate


class GaussianProcessPreset(Enum):
    """Available Gaussian process surrogate presets."""

    BAYBE = "BAYBE"
    """Recreates the default settings of the Gaussian process surrogate class."""


def make_gp_from_preset(preset: GaussianProcessPreset) -> GaussianProcessSurrogate:
    """Create a :class:`GaussianProcessSurrogate` from a :class:`GaussianProcessPreset."""  # noqa: E501
    from baybe.surrogates.gaussian_process.core import GaussianProcessSurrogate

    if preset is GaussianProcessPreset.BAYBE:
        return GaussianProcessSurrogate()

    raise ValueError(
        f"Unknown '{GaussianProcessPreset.__name__}' with name '{preset.name}'."
    )



"""Validation utilities."""

from __future__ import annotations

import math
from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Any

import numpy as np
import pandas as pd
from attrs import Attribute

if TYPE_CHECKING:
    from baybe.parameters.base import Parameter
    from baybe.targets.base import Target


def validate_not_nan(self: Any, attribute: Attribute, value: Any) -> None:
    """Attrs-compatible validator to forbid 'nan' values."""
    if isinstance(value, float) and math.isnan(value):
        raise ValueError(
            f"The value passed to attribute '{attribute.name}' of class "
            f"'{self.__class__.__name__}' cannot be 'nan'."
        )


def _make_restricted_float_validator(
    allow_nan: bool, allow_inf: bool
) -> Callable[[Any, Attribute, Any], None]:
    """Make an attrs-compatible validator for restricted floats.

    Args:
        allow_nan: If False, validated values cannot be 'nan'.
        allow_inf: If False, validated values cannot be 'inf' or '-inf'.

    Raises:
        ValueError: If no float range restriction is in place.

    Returns:
        The validator.
    """
    if allow_nan and allow_inf:
        raise ValueError(
            "The requested validator would not restrict the float range. "
            "Hence, you can use `attrs.validators.instance_of(float)` instead."
        )

    def validator(self: Any, attribute: Attribute, value: Any) -> None:
        if not isinstance(value, float):
            raise ValueError(
                f"Values assigned to attribute '{attribute.name}' of class "
                f"'{self.__class__.__name__}' must be of type 'float'. "
                f"Given: {value} (type: {type(value)})"
            )
        if not allow_inf and math.isinf(value):
            raise ValueError(
                f"Values assigned to attribute '{attribute.name}' of class "
                f"'{self.__class__.__name__}' cannot be 'inf' or '-inf'."
            )
        if not allow_nan and math.isnan(value):
            raise ValueError(
                f"Values assigned to attribute '{attribute.name}' of class "
                f"'{self.__class__.__name__}' cannot be 'nan'."
            )

    return validator


finite_float = _make_restricted_float_validator(allow_nan=False, allow_inf=False)
"""Validator for finite (i.e., non-nan and non-infinite) floats."""

non_nan_float = _make_restricted_float_validator(allow_nan=False, allow_inf=True)
"""Validator for non-nan floats."""

non_inf_float = _make_restricted_float_validator(allow_nan=True, allow_inf=False)
"""Validator for non-infinite floats."""


def validate_target_input(data: pd.DataFrame, targets: Iterable[Target]) -> None:
    """Validate input dataframe columns corresponding to targets.

    Args:
        data: The input dataframe to be validated.
        targets: The allowed targets.

    Raises:
        ValueError: If the data is empty.
        ValueError: If the data misses columns for a target.
        TypeError: If any numerical target data contain non-numeric values.
        ValueError: If any binary target data contain values not part of the targets'
            allowed values or NaN.
    """
    from baybe.targets import BinaryTarget, NumericalTarget

    if data.empty:
        raise ValueError("The provided input dataframe cannot be empty.")

    if missing := {t.name for t in targets}.difference(data.columns):
        raise ValueError(
            f"The input dataframe is missing columns for the following targets: "
            f"{missing}"
        )

    for t in targets:
        if isinstance(t, NumericalTarget):
            if data[t.name].dtype.kind not in "iufb":
                raise TypeError(
                    f"The numerical target '{t.name}' has non-numeric entries in the "
                    f"provided dataframe."
                )
        elif isinstance(t, BinaryTarget):
            allowed = {t.failure_value, t.success_value, np.nan}
            if invalid := set(data[t.name].unique()) - allowed:
                raise ValueError(
                    f"The binary target '{t.name}' has invalid entries {invalid} "
                    f"in the provided dataframe. Allowed values are: {allowed}."
                )


def validate_parameter_input(
    data: pd.DataFrame,
    parameters: Iterable[Parameter],
    numerical_measurements_must_be_within_tolerance: bool = False,
) -> None:
    """Validate input dataframe columns corresponding to parameters.

    Args:
        data: The input dataframe to be validated.
        parameters: The allowed parameters.
        numerical_measurements_must_be_within_tolerance: If ``True``, numerical
            parameter values must match to parameter values within the
            parameter-specific tolerance.

    Raises:
        ValueError: If the data is empty.
        ValueError: If the data misses columns for a parameter.
        ValueError: If a parameter contains NaN.
        TypeError: If a parameter contains non-numeric values.
    """
    if data.empty:
        raise ValueError("The provided input dataframe cannot be empty.")

    if missing := {p.name for p in parameters}.difference(data.columns):
        raise ValueError(
            f"The input dataframe is missing columns for the following parameters: "
            f"{missing}"
        )

    for p in parameters:
        if data[p.name].isna().any():
            raise ValueError(
                f"The parameter '{p.name}' has missing values in the provided "
                f"dataframe."
            )
        if p.is_numerical and (data[p.name].dtype.kind not in "iufb"):
            raise TypeError(
                f"The numerical parameter '{p.name}' has non-numeric entries in the "
                f"provided dataframe."
            )

        # Check if all rows have valid inputs matching allowed parameter values
        for ind, row in data.iterrows():
            valid = True
            if p.is_numerical:
                if numerical_measurements_must_be_within_tolerance:
                    valid &= p.is_in_range(row[p.name])
            else:
                valid &= p.is_in_range(row[p.name])
            if not valid:
                raise ValueError(
                    f"Input data on row with the index {row.name} has invalid "
                    f"values in parameter '{p.name}'. "
                    f"For categorical parameters, values need to exactly match a "
                    f"valid choice defined in your config. "
                    f"For numerical parameters, a match is accepted only if "
                    f"the input value is within the specified tolerance/range. Set "
                    f"the flag 'numerical_measurements_must_be_within_tolerance' "
                    f"to 'False' to disable this behavior."
                )



"""Temporary namespace for backward compatibility."""

import warnings

from baybe.recommenders.pure.bayesian.botorch import BotorchRecommender


def SequentialGreedyRecommender(*args, **kwargs) -> BotorchRecommender:
    """A :class:`baybe.recommenders.pure.bayesian.botorch.BotorchRecommender` alias for backward compatibility."""  # noqa: D401, E501
    warnings.warn(
        f"The class `SequentialGreedyRecommender` has been deprecated and will be "
        f"removed in a future version. "
        f"Please use `{BotorchRecommender.__name__}(sequential_continuous=True)` "
        f"instead.",
        DeprecationWarning,
    )
    return BotorchRecommender(*args, **kwargs, sequential_continuous=True)



"""Benchmarking module for performance tracking."""

from benchmarks.definition import (
    Benchmark,
    BenchmarkSettings,
    ConvergenceBenchmark,
    ConvergenceBenchmarkSettings,
)
from benchmarks.result import Result

__all__ = [
    "Benchmark",
    "BenchmarkSettings",
    "ConvergenceBenchmark",
    "ConvergenceBenchmarkSettings",
    "Result",
]



"""Default preset for Gaussian process surrogates."""

from __future__ import annotations

import gc
from typing import TYPE_CHECKING

import numpy as np
from attrs import define
from typing_extensions import override

from baybe.kernels.basic import MaternKernel
from baybe.kernels.composite import ScaleKernel
from baybe.parameters import TaskParameter
from baybe.priors.basic import GammaPrior
from baybe.surrogates.gaussian_process.kernel_factory import KernelFactory

if TYPE_CHECKING:
    from torch import Tensor

    from baybe.kernels.base import Kernel
    from baybe.searchspace.core import SearchSpace

# Boundaries for low and high dimension limits
_DIM_LIMITS = (8, 75)


@define
class DefaultKernelFactory(KernelFactory):
    """A factory providing the default kernel for Gaussian process surrogates.

    This is taking the low and high dimensional limits of
    :class:`baybe.surrogates.gaussian_process.presets.edbo.EDBOKernelFactory`
    and interpolates the prior moments linearly between them.
    """

    @override
    def __call__(
        self, searchspace: SearchSpace, train_x: Tensor, train_y: Tensor
    ) -> Kernel:
        effective_dims = train_x.shape[-1] - len(
            [p for p in searchspace.parameters if isinstance(p, TaskParameter)]
        )

        # Interpolate prior moments linearly between low D and high D regime
        # The high D regime itself is the average of the EDBO OHE and Mordred regime
        # Values outside the dimension limits will get the border value assigned
        lengthscale_prior = GammaPrior(
            np.interp(effective_dims, _DIM_LIMITS, [1.2, 2.5]),
            np.interp(effective_dims, _DIM_LIMITS, [1.1, 0.55]),
        )
        lengthscale_initial_value = np.interp(effective_dims, _DIM_LIMITS, [0.2, 6.0])
        outputscale_prior = GammaPrior(
            np.interp(effective_dims, _DIM_LIMITS, [5.0, 3.5]),
            np.interp(effective_dims, _DIM_LIMITS, [0.5, 0.15]),
        )
        outputscale_initial_value = np.interp(effective_dims, _DIM_LIMITS, [8.0, 15.0])

        return ScaleKernel(
            MaternKernel(
                nu=2.5,
                lengthscale_prior=lengthscale_prior,
                lengthscale_initial_value=lengthscale_initial_value,
            ),
            outputscale_prior=outputscale_prior,
            outputscale_initial_value=outputscale_initial_value,
        )


def _default_noise_factory(
    searchspace: SearchSpace, train_x: Tensor, train_y: Tensor
) -> tuple[GammaPrior, float]:
    """Create the default noise settings for the Gaussian process surrogate.

    This is taking the low and high dimensional limits of
    :func:`baybe.surrogates.gaussian_process.presets.edbo._edbo_noise_factory`
    and interpolates the prior moments linearly between them.
    """
    # TODO: Replace this function with a proper likelihood factory

    # Interpolate prior moments linearly between low D and high D regime
    # The high D regime itself is the average of the EDBO OHE and Mordred regime
    # Values outside the dimension limits will get the border value assigned
    effective_dims = train_x.shape[-1] - len(
        [p for p in searchspace.parameters if isinstance(p, TaskParameter)]
    )
    return (
        GammaPrior(
            np.interp(effective_dims, _DIM_LIMITS, [1.05, 1.5]),
            np.interp(effective_dims, _DIM_LIMITS, [0.5, 0.1]),
        ),
        np.interp(effective_dims, _DIM_LIMITS, [0.1, 5.0]).item(),
    )


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Benchmark result metadata."""

from datetime import datetime, timedelta

import git
from attrs import define, field
from attrs.validators import instance_of, optional
from cattrs.gen import make_dict_unstructure_fn

from benchmarks.serialization import BenchmarkSerialization, converter


@define(frozen=True)
class ResultMetadata(BenchmarkSerialization):
    """The metadata of a benchmark result."""

    start_datetime: datetime = field(validator=instance_of(datetime))
    """The start datetime of the benchmark."""

    duration: timedelta = field(validator=instance_of(timedelta))
    """The time it took to complete the benchmark."""

    commit_hash: str = field(validator=instance_of(str), init=False)
    """The commit hash of the used BayBE code."""

    latest_baybe_tag: str = field(validator=instance_of(str), init=False)
    """The latest BayBE tag reachable in the ancestor commit history."""

    branch: str | None = field(validator=optional(instance_of(str)), init=False)
    """The branch checked out during benchmark execution."""

    @branch.default
    def _default_branch(self) -> str | None:
        """Set the current checkout branch."""
        repo = git.Repo(search_parent_directories=True)
        try:
            current_branch = repo.active_branch.name
            return current_branch
        except TypeError:
            return None

    @commit_hash.default
    def _default_commit_hash(self) -> str:
        """Extract the git commit hash."""
        repo = git.Repo(search_parent_directories=True)
        sha = repo.head.object.hexsha
        return sha

    @latest_baybe_tag.default
    def _default_latest_baybe_tag(self) -> str:
        """Extract the latest reachable BayBE tag."""
        repo = git.Repo(search_parent_directories=True)
        latest_tag = repo.git.describe(tags=True, abbrev=0)
        return latest_tag


# Register un-/structure hooks
converter.register_unstructure_hook(
    ResultMetadata,
    make_dict_unstructure_fn(
        ResultMetadata, converter, _cattrs_include_init_false=True
    ),
)



"""Scaling tests."""

import math
from unittest.mock import Mock

import pytest
import torch
from botorch.models.transforms.input import InputStandardize, InputTransform, Normalize

from baybe.utils.scaling import ColumnTransformer


def test_column_transformer():
    """Basic test that validates the transformation of a ColumnTransformer."""
    # Define test input
    base = torch.stack([-1 * torch.ones(5), torch.ones(5)])
    x_train = 20 * base
    x_test = 10 * base

    # Create and fit transformer
    mapping = {
        (0, 2): Normalize(2),
        (1, 3): InputStandardize(2),
    }
    transformer = ColumnTransformer(mapping)
    transformer.fit(x_train)

    # Transform data
    y_train = transformer.transform(x_train)
    y_test = transformer.transform(x_test)

    # Expected results
    s = 1 / math.sqrt(2)
    target_train = torch.tensor(
        [
            [0.0, -s, 0.0, -s, -20.0],
            [1.0, s, 1.0, s, 20.0],
        ]
    )
    target_test = torch.tensor(
        [
            [0.25, -s / 2, 0.25, -s / 2, -10.0],
            [0.75, s / 2, 0.75, s / 2, 10.0],
        ]
    )

    assert torch.allclose(y_train, target_train)
    assert torch.allclose(y_test, target_test)


def test_non_disjoint_column_mapping():
    """Creating a column transformer with non-disjoint columns raises an error."""
    t = Mock(InputTransform)
    mapping = {(0, 1): t, (1, 2): t}
    with pytest.raises(ValueError, match="are not disjoint"):
        ColumnTransformer(mapping)



"""Test serialization of objectives."""

import hypothesis.strategies as st
import pytest
from hypothesis import given
from pytest import param

from baybe.objectives.base import Objective
from tests.hypothesis_strategies.objectives import (
    desirability_objectives,
    pareto_objectives,
    single_target_objectives,
)


@pytest.mark.parametrize(
    "objective_strategy",
    [
        param(single_target_objectives(), id="SingleTargetObjective"),
        param(desirability_objectives(), id="DesirabilityObjective"),
        param(pareto_objectives(), id="ParetoObjective"),
    ],
)
@given(data=st.data())
def test_objective_roundtrip(objective_strategy, data):
    """A serialization roundtrip yields an equivalent object."""
    objective = data.draw(objective_strategy)
    string = objective.to_json()
    objective2 = Objective.from_json(string)
    assert objective == objective2, (objective, objective2)



"""Base classes for all parameters."""

from __future__ import annotations

import gc
from abc import ABC, abstractmethod
from functools import cached_property, partial
from typing import TYPE_CHECKING, Any, ClassVar

import cattrs
import pandas as pd
from attrs import define, field
from attrs.converters import optional as optional_c
from attrs.validators import instance_of, min_len
from typing_extensions import override

from baybe.parameters.enum import ParameterEncoding
from baybe.serialization import (
    SerialMixin,
    converter,
    get_base_structure_hook,
    unstructure_base,
)
from baybe.utils.basic import to_tuple

if TYPE_CHECKING:
    from baybe.searchspace.continuous import SubspaceContinuous
    from baybe.searchspace.core import SearchSpace
    from baybe.searchspace.discrete import SubspaceDiscrete

# TODO: Reactive slots in all classes once cached_property is supported:
#   https://github.com/python-attrs/attrs/issues/164


@define(frozen=True, slots=False)
class Parameter(ABC, SerialMixin):
    """Abstract base class for all parameters.

    Stores information about the type, range, constraints, etc. and handles in-range
    checks, transformations etc.
    """

    # class variables
    is_numerical: ClassVar[bool]
    """Class variable encoding whether this parameter is numeric."""

    # object variables
    name: str = field(validator=(instance_of(str), min_len(1)))
    """The name of the parameter"""

    @abstractmethod
    def is_in_range(self, item: Any) -> bool:
        """Return whether an item is within the parameter range.

        Args:
            item: The item to be checked.

        Returns:
            ``True`` if the item is within the parameter range, ``False`` otherwise.
        """

    @override
    def __str__(self) -> str:
        return str(self.summary())

    @property
    def is_continuous(self) -> bool:
        """Boolean indicating if this is a continuous parameter."""
        return isinstance(self, ContinuousParameter)

    @property
    def is_discrete(self) -> bool:
        """Boolean indicating if this is a discrete parameter."""
        return isinstance(self, DiscreteParameter)

    @property
    @abstractmethod
    def comp_rep_columns(self) -> tuple[str, ...]:
        """The columns spanning the computational representation."""

    def to_searchspace(self) -> SearchSpace:
        """Create a one-dimensional search space from the parameter."""
        from baybe.searchspace.core import SearchSpace

        return SearchSpace.from_parameter(self)

    @abstractmethod
    def summary(self) -> dict:
        """Return a custom summarization of the parameter."""


@define(frozen=True, slots=False)
class DiscreteParameter(Parameter, ABC):
    """Abstract class for discrete parameters."""

    # class variables
    encoding: ParameterEncoding | None = field(init=False, default=None)
    """An optional encoding for the parameter."""

    @property
    @abstractmethod
    def values(self) -> tuple:
        """The values the parameter can take."""

    @property
    def active_values(self) -> tuple:
        """The values that are considered for recommendation."""
        return self.values

    @cached_property
    @abstractmethod
    def comp_df(self) -> pd.DataFrame:
        # TODO: Should be renamed to `comp_rep`
        """Return the computational representation of the parameter."""

    @override
    @property
    def comp_rep_columns(self) -> tuple[str, ...]:
        return tuple(self.comp_df.columns)

    def to_subspace(self) -> SubspaceDiscrete:
        """Create a one-dimensional search space from the parameter."""
        from baybe.searchspace.discrete import SubspaceDiscrete

        return SubspaceDiscrete.from_parameter(self)

    @override
    def is_in_range(self, item: Any) -> bool:
        return item in self.values

    def transform(self, series: pd.Series, /) -> pd.DataFrame:
        """Transform parameter values to computational representation.

        Args:
            series: The parameter values in experimental representation to be
                transformed.

        Returns:
            A series containing the transformed values. The series name matches
            that of the input.
        """
        if self.encoding:
            # replace each label with the corresponding encoding
            transformed = pd.merge(
                left=series.rename("Labels").to_frame(),
                left_on="Labels",
                right=self.comp_df,
                right_index=True,
                how="left",
            ).drop(columns="Labels")
        else:
            transformed = series.to_frame()

        return transformed

    @override
    def summary(self) -> dict:
        param_dict = dict(
            Name=self.name,
            Type=self.__class__.__name__,
            nValues=len(self.values),
            Encoding=self.encoding,
        )
        return param_dict


@define(frozen=True, slots=False)
class _DiscreteLabelLikeParameter(DiscreteParameter, ABC):
    """Abstract class for discrete label-like parameters.

    In general, these are parameters with non-numerical experimental representations.
    """

    # class variables
    is_numerical: ClassVar[bool] = False
    # See base class.

    # object variables
    _active_values: tuple[str | bool, ...] | None = field(
        default=None,
        converter=optional_c(to_tuple),
        kw_only=True,
        alias="active_values",
    )
    """Optional labels identifying the ones which should be actively recommended."""

    @override
    @property
    def active_values(self) -> tuple[str | bool, ...]:
        if self._active_values is None:
            return self.values

        return self._active_values

    @_active_values.validator
    def _validate_active_values(  # noqa: DOC101, DOC103
        self, _: Any, content: tuple[str | bool, ...]
    ) -> None:
        """Validate the active parameter values.

        If no such list is provided, no validation is being performed. In particular,
        the errors listed below are only relevant if the ``values`` list is provided.

        Raises:
            ValueError: If an empty active parameters list is provided.
            ValueError: If the active parameter values are not unique.
            ValueError: If not all active values are valid parameter choices.
        """
        if content is None:
            return

        if len(content) == 0:
            raise ValueError(
                "If an active parameters list is provided, it must not be empty."
            )
        if len(set(content)) != len(content):
            raise ValueError("The active parameter values must be unique.")
        if not all(v in self.values for v in content):
            raise ValueError(
                f"All active values must be valid parameter choices from: "
                f"{self.values}, provided: {content}"
            )

    @override
    def summary(self) -> dict:
        return {**super().summary(), "nActiveValues": len(self.active_values)}


@define(frozen=True, slots=False)
class ContinuousParameter(Parameter):
    """Abstract class for continuous parameters."""

    def to_subspace(self) -> SubspaceContinuous:
        """Create a one-dimensional search space from the parameter."""
        from baybe.searchspace.continuous import SubspaceContinuous

        return SubspaceContinuous.from_parameter(self)


# Register (un-)structure hooks
_overrides = {
    "_values": cattrs.override(rename="values"),
    "_active_values": cattrs.override(rename="active_values"),
}
# FIXME[typing]: https://github.com/python/mypy/issues/4717
converter.register_structure_hook(
    Parameter,
    get_base_structure_hook(Parameter, overrides=_overrides),  # type: ignore
)
converter.register_unstructure_hook(
    Parameter, partial(unstructure_base, overrides=_overrides)
)

# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Deserialization tests.

The purpose of these tests is to ensure that deserialization works for all possible ways
in which type information can be provided.

NOTE: The tests are based on `AcquisitionFunction` simply because the class provides the
    `abbreviation` variable that is necessary for the tests.
"""

import pytest
from pytest import param

from baybe.acquisition.acqfs import UpperConfidenceBound
from baybe.acquisition.base import AcquisitionFunction


@pytest.mark.parametrize(
    ("cls", "dct"),
    [
        param(
            AcquisitionFunction,
            {"type": "UpperConfidenceBound", "beta": 1337},
            id="base-full",
        ),
        param(
            AcquisitionFunction,
            {"type": "UCB", "beta": 1337},
            id="base-abbreviation",
        ),
        param(
            UpperConfidenceBound,
            {"type": "UpperConfidenceBound", "beta": 1337},
            id="subclass-full",
        ),
        param(
            UpperConfidenceBound,
            {"type": "UCB", "beta": 1337},
            id="subclass-abbreviation",
        ),
        param(
            UpperConfidenceBound,
            {"beta": 1337},
            id="subclass-without",
        ),
    ],
)
def test_valid_deserialization(cls, dct):
    """Serialization is possible from concrete class and from base class.

    Concrete class works with: full type name, type abbreviation, and without type.
    Base class works with: full type name and type abbreviation.
    """
    UpperConfidenceBound(beta=1337) == cls.from_dict(dct)


@pytest.mark.parametrize(
    ("cls", "dct", "err", "msg"),
    [
        param(
            AcquisitionFunction,
            {"beta": 1337},
            KeyError,
            "type",
            id="missing-type",
        ),
        param(
            UpperConfidenceBound,
            {"type": "wrong", "beta": 1337},
            ValueError,
            "'UpperConfidenceBound' .* does not match .* 'wrong'",
            id="inconsistent-type",
        ),
    ],
)
def test_invalid_deserialization(cls, dct, err, msg):
    """Incorrect type information throws the correct error."""
    with pytest.raises(err, match=msg):
        cls.from_dict(dct)



"""Validation tests for objective."""

from unittest.mock import Mock

import pytest
from cattrs import IterableValidationError
from pytest import param

from baybe.objectives import ParetoObjective
from baybe.objectives.desirability import DesirabilityObjective
from baybe.objectives.single import SingleTargetObjective
from baybe.targets.base import Target
from baybe.targets.numerical import NumericalTarget

t1 = NumericalTarget("t1", mode="MAX", bounds=(0, 1))
t2 = NumericalTarget("t2", mode="MAX", bounds=(0, 1))
t3 = NumericalTarget("unnormalized", mode="MAX")
t_mock = Mock(spec=Target)


@pytest.mark.parametrize(
    "target",
    [
        param(None, id="none"),
        param("no_target", id="string"),
    ],
)
def test_invalid_target(target):
    """Providing an invalid target object raises an exception."""
    with pytest.raises(TypeError):
        SingleTargetObjective(target)


@pytest.mark.parametrize(
    ("targets", "error"),
    [
        param(None, TypeError, id="none"),
        param([t1, "t2"], TypeError, id="wrong_type"),
        param([t1], ValueError, id="too_short"),
        param([t1, t1], ValueError, id="duplicate_names"),
        param([t1, t3], ValueError, id="unnormalized"),
        param([t1, t_mock], TypeError, id="unsupported_subclass"),
    ],
)
def test_invalid_targets(targets, error):
    """Providing invalid target objects raises an exception."""
    with pytest.raises(error):
        DesirabilityObjective(targets)


@pytest.mark.parametrize(
    ("weights", "error"),
    [
        param(None, TypeError, id="none"),
        param([1.0, "abc"], IterableValidationError, id="wrong_type"),
        param([1.0, -1.0], ValueError, id="negative"),
        param([1.0, 0.0], ValueError, id="zero"),
        param([1.0, float("inf")], ValueError, id="inf"),
        param([1.0, float("nan")], ValueError, id="nan"),
        param([1.0], ValueError, id="wrong_length"),
    ],
)
def test_invalid_weights(weights, error):
    """Providing invalid weights raises an exception."""
    with pytest.raises(error):
        DesirabilityObjective([t1, t2], weights)


@pytest.mark.parametrize(
    "scalarizer",
    [
        param(None, id="none"),
        param("invalid", id="non_existing"),
    ],
)
def test_invalid_scalarizer(scalarizer):
    """Providing an invalid scalarizer raises an exception."""
    with pytest.raises(ValueError):
        DesirabilityObjective([t1, t2], scalarizer=scalarizer)


@pytest.mark.parametrize(
    "target",
    [
        param(NumericalTarget("t2", mode="MAX", bounds=(0, 1)), id="max_bounded"),
        param(NumericalTarget("t2", mode="MIN", bounds=(0, 1)), id="min_bounded"),
    ],
)
def test_pareto_transformed_targets(target):
    """Providing a bounded/transformed MIX/MAX target raises an exception."""
    with pytest.raises(ValueError, match="does not support transforms for"):
        ParetoObjective([NumericalTarget("t1", mode="MAX"), target])



## Optimizing a Custom Black-Box Function

# This example demonstrates how to optimize a custom black-box function:
# * We create a black-box callable and define the corresponding optimization scope,
# * set up optimization strategies,
# * and compare the resulting trajectories.

### Imports

import os

import pandas as pd
import seaborn as sns

from baybe import Campaign
from baybe.parameters.numerical import NumericalContinuousParameter
from baybe.recommenders import RandomRecommender
from baybe.searchspace import SearchSpace
from baybe.simulation import simulate_scenarios
from baybe.targets import NumericalTarget
from examples.utils import create_example_plots

### Settings

# Before we start, let us collect a few general settings for the example:

SMOKE_TEST = "SMOKE_TEST" in os.environ

BATCH_SIZE = 1
N_MC_ITERATIONS = 2 if SMOKE_TEST else 20
N_DOE_ITERATIONS = 2 if SMOKE_TEST else 30
DIMENSION = 2 if SMOKE_TEST else 10
BOUNDS = (-1, 1)

### Defining the Optimization Problem

# Now, we can define the scope of our optimization problem. Our goal is to optimize
# a high-dimensional quadratic function on a bounded input domain. We first define
# the corresponding inputs and output of the function:

parameters = [
    NumericalContinuousParameter(name=f"x_{k}", bounds=BOUNDS) for k in range(DIMENSION)
]
target = NumericalTarget(name="Target", mode="MIN")


# Based on the above, we construct the black-box callable to be optimized, which
# provides the lookup mechanism for closing the optimization loop:


def blackbox(df: pd.DataFrame, /) -> pd.DataFrame:
    """A callable whose internal logic is unknown to the algorithm."""
    return (df[[p.name for p in parameters]] ** 2).sum(axis=1).to_frame(target.name)


# What remains is to construct the search space and objective for the optimization:

searchspace = SearchSpace.from_product(parameters=parameters)
objective = target.to_objective()

### Creating the Campaigns

# We consider two optimization scenarios, each represented by its own campaign:
# * Optimization using the default recommender
# * A baseline using randomly generated recommendations

default_campaign = Campaign(
    searchspace=searchspace,
    objective=objective,
)
random_campaign = Campaign(
    searchspace=searchspace,
    objective=objective,
    recommender=RandomRecommender(),
)

### Running the Simulation Loop

# Next, we simulate both scenarios using the
# {func}`~baybe.simulation.scenarios.simulate_scenarios` utility,
# which automatically executes several Monte Carlo simulations for each campaign:

scenarios = {
    "Default Recommender": default_campaign,
    "Random Recommender": random_campaign,
}
results = simulate_scenarios(
    scenarios,
    blackbox,
    batch_size=BATCH_SIZE,
    n_doe_iterations=N_DOE_ITERATIONS,
    n_mc_iterations=N_MC_ITERATIONS,
)

### Plotting the Results

# Finally, we compare the trajectories of the campaigns:

ax = sns.lineplot(
    data=results,
    marker="o",
    markersize=10,
    x="Num_Experiments",
    y="Target_CumBest",
    hue="Scenario",
)
create_example_plots(ax=ax, base_name="custom_blackbox")



"""Adapter functionality for making BayBE surrogates BoTorch-ready."""

from collections.abc import Callable
from typing import Any

from botorch.models.gpytorch import Model
from botorch.posteriors import Posterior
from torch import Tensor

from baybe.surrogates.base import Surrogate


class AdapterModel(Model):
    """A BoTorch model that uses a BayBE surrogate model for posterior computation.

    Can be used, for example, as an adapter layer for making a BayBE
    surrogate model usable in conjunction with BoTorch acquisition functions.

    Args:
        surrogate: The internal surrogate model.
    """

    def __init__(self, surrogate: Surrogate):
        super().__init__()
        self._surrogate = surrogate

    @property
    def num_outputs(self) -> int:
        if (objective := self._surrogate._objective) is None:
            raise RuntimeError(
                "The model needs to be fitted to a specific context first."
            )
        return objective.n_outputs

    def posterior(
        self,
        X: Tensor,
        output_indices: list[int] | None = None,
        observation_noise: bool = False,
        posterior_transform: Callable[[Posterior], Posterior] | None = None,
        **kwargs: Any,
    ) -> Posterior:
        if (
            (output_indices is not None)
            or observation_noise
            or (posterior_transform is not None)
        ):
            raise NotImplementedError(
                "The optional model posterior arguments are not yet implemented."
            )
        return self._surrogate._posterior_comp(X)



"""Parameter-related enumerations."""

from enum import Enum


class ParameterEncoding(Enum):
    """Generic base class for all parameter encodings."""


class CategoricalEncoding(ParameterEncoding):
    """Available encodings for categorical parameters."""

    OHE = "OHE"
    """One-hot encoding."""

    INT = "INT"
    """Integer encoding."""


class CustomEncoding(ParameterEncoding):
    """Available encodings for custom parameters."""

    CUSTOM = "CUSTOM"
    """User-defined encoding."""


class SubstanceEncoding(ParameterEncoding):
    """Available encodings for substance parameters from `scikit-fingerprints`_ package.

    .. _scikit-fingerprints: https://scikit-fingerprints.github.io/scikit-fingerprints/
    """

    ATOMPAIR = "ATOMPAIR"
    """:class:`skfp.fingerprints.AtomPairFingerprint`"""

    AUTOCORR = "AUTOCORR"
    """:class:`skfp.fingerprints.AutocorrFingerprint`"""

    AVALON = "AVALON"
    """:class:`skfp.fingerprints.AvalonFingerprint`"""

    BCUT2D = "BCUT2D"
    """:class:`skfp.fingerprints.BCUT2DFingerprint`"""

    E3FP = "E3FP"
    """:class:`skfp.fingerprints.E3FPFingerprint`"""

    ECFP = "ECFP"
    """:class:`skfp.fingerprints.ECFPFingerprint`"""

    ELECTROSHAPE = "ELECTROSHAPE"
    """:class:`skfp.fingerprints.ElectroShapeFingerprint`"""

    MORGAN_FP = "MORGAN_FP"
    """
    Deprecated! Uses :class:`skfp.fingerprints.ECFPFingerprint` with ``fp_size=1024``
    and ``radius=4``.
    """

    ERG = "ERG"
    """:class:`skfp.fingerprints.ERGFingerprint`"""

    ESTATE = "ESTATE"
    """:class:`skfp.fingerprints.EStateFingerprint`"""

    FUNCTIONALGROUPS = "FUNCTIONALGROUPS"
    """:class:`skfp.fingerprints.FunctionalGroupsFingerprint`"""

    GETAWAY = "GETAWAY"
    """:class:`skfp.fingerprints.GETAWAYFingerprint`"""

    GHOSECRIPPEN = "GHOSECRIPPEN"
    """:class:`skfp.fingerprints.GhoseCrippenFingerprint`"""

    KLEKOTAROTH = "KLEKOTAROTH"
    """:class:`skfp.fingerprints.KlekotaRothFingerprint`"""

    LAGGNER = "LAGGNER"
    """:class:`skfp.fingerprints.LaggnerFingerprint`"""

    LAYERED = "LAYERED"
    """:class:`skfp.fingerprints.LayeredFingerprint`"""

    LINGO = "LINGO"
    """:class:`skfp.fingerprints.LingoFingerprint`"""

    MACCS = "MACCS"
    """:class:`skfp.fingerprints.MACCSFingerprint`"""

    MAP = "MAP"
    """:class:`skfp.fingerprints.MAPFingerprint`"""

    MHFP = "MHFP"
    """:class:`skfp.fingerprints.MHFPFingerprint`"""

    MORSE = "MORSE"
    """:class:`skfp.fingerprints.MORSEFingerprint`"""

    MQNS = "MQNS"
    """:class:`skfp.fingerprints.MQNsFingerprint`"""

    MORDRED = "MORDRED"
    """:class:`skfp.fingerprints.MordredFingerprint`"""

    PATTERN = "PATTERN"
    """:class:`skfp.fingerprints.PatternFingerprint`"""

    PHARMACOPHORE = "PHARMACOPHORE"
    """:class:`skfp.fingerprints.PharmacophoreFingerprint`"""

    PHYSIOCHEMICALPROPERTIES = "PHYSIOCHEMICALPROPERTIES"
    """:class:`skfp.fingerprints.PhysiochemicalPropertiesFingerprint`"""

    PUBCHEM = "PUBCHEM"
    """:class:`skfp.fingerprints.PubChemFingerprint`"""

    RDF = "RDF"
    """:class:`skfp.fingerprints.RDFFingerprint`"""

    RDKIT = "RDKIT"
    """Deprecated! Uses :class:`skfp.fingerprints.RDKit2DDescriptors`."""

    RDKITFINGERPRINT = "RDKITFINGERPRINT"
    """:class:`skfp.fingerprints.RDKitFingerprint`"""

    RDKIT2DDESCRIPTORS = "RDKIT2DDESCRIPTORS"
    """:class:`skfp.fingerprints.RDKit2DDescriptorsFingerprint`"""

    SECFP = "SECFP"
    """:class:`skfp.fingerprints.SECFPFingerprint`"""

    TOPOLOGICALTORSION = "TOPOLOGICALTORSION"
    """:class:`skfp.fingerprints.TopologicalTorsionFingerprint`"""

    USR = "USR"
    """:class:`skfp.fingerprints.USRFingerprint`"""

    USRCAT = "USRCAT"
    """:class:`skfp.fingerprints.USRCATFingerprint`"""

    VSA = "VSA"
    """:class:`skfp.fingerprints.VSAFingerprint`"""

    WHIM = "WHIM"
    """:class:`skfp.fingerprints.WHIMFingerprint`"""



"""Tests for parameter utilities."""

from unittest.mock import Mock

import pytest
from pytest import param

from baybe.parameters import NumericalContinuousParameter
from baybe.parameters.utils import activate_parameter
from baybe.utils.interval import Interval


def mirror_interval(interval: Interval) -> Interval:
    """Return an interval copy mirrored around the origin."""
    return Interval(lower=-interval.upper, upper=-interval.lower)


@pytest.mark.parametrize("mirror", [False, True], ids=["regular", "mirrored"])
@pytest.mark.parametrize(
    ("bounds", "thresholds", "expected_result"),
    [
        # Valid case: When parameter lower/upper bounds spread in both
        # negative/positive directions.
        param(
            Interval(lower=-1.0, upper=1.0),
            Interval(lower=-0.5, upper=0.5),
            Interval(lower=-1.0, upper=1.0),
            id="valid_thresholds_in_bounds",
        ),
        # Valid case: When one parameter bound is zero.
        param(
            Interval(lower=-1.0, upper=0.0),
            Interval(lower=-0.5, upper=0.0),
            Interval(lower=-1.0, upper=-0.5),
            id="valid_thresholds_in_bounds_one_side_match_with_zero_threshold",
        ),
        # Invalid case: Activating a parameter requires a valid inactive region to
        # start with.
        param(
            None,
            Interval(lower=0.5, upper=1.0),
            ValueError,
            id="invalid_inactive_region",
        ),
        # Invalid cases related to NotImplementedError. To test whether we can
        # correctly catch different NotImplementedError cases, both nonzero/zero
        # thresholds are considered when necessary.
        param(
            Interval(lower=-1.0, upper=1.0),
            Interval(lower=-0.5, upper=1.0),
            NotImplementedError,
            id="invalid_thresholds_in_bounds_one_side_match_with_nonzero_threshold",
        ),
        param(
            Interval(lower=-1.0, upper=1.0),
            Interval(lower=-1.0, upper=1.0),
            NotImplementedError,
            id="invalid_bounds_on_thresholds_with_nonzero_threshold",
        ),
        param(
            Interval(lower=0.0, upper=1.0),
            Interval(lower=0.0, upper=1.0),
            NotImplementedError,
            id="invalid_bounds_on_thresholds_with_zero_threshold",
        ),
        param(
            Interval(lower=-1.0, upper=1.0),
            Interval(lower=-1.5, upper=1.5),
            NotImplementedError,
            id="invalid_bounds_in_thresholds",
        ),
        param(
            Interval(lower=-1.0, upper=1.0),
            Interval(lower=-1.5, upper=1.0),
            NotImplementedError,
            id="invalid_bounds_in_thresholds_one_side_match_with_nonzero_threshold",
        ),
        param(
            Interval(lower=-1.0, upper=0.0),
            Interval(lower=-1.5, upper=0.0),
            NotImplementedError,
            id="invalid_bounds_in_thresholds_one_side_match_with_zero_threshold",
        ),
        param(
            Interval(lower=-0.5, upper=1.0),
            Interval(lower=-1.0, upper=0.5),
            NotImplementedError,
            id="invalid_bounds_intersected_with_thresholds",
        ),
        param(
            Interval(lower=0.5, upper=1.0),
            Interval(lower=-1.0, upper=0.5),
            NotImplementedError,
            id="invalid_bounds_intersected_with_thresholds_on_nonzero_one_point",
        ),
        param(
            Interval(lower=0.0, upper=1.0),
            Interval(lower=-1.0, upper=0.0),
            NotImplementedError,
            id="invalid_bounds_intersected_with_thresholds_on_zero_one_point",
        ),
        param(
            Interval(lower=0.5, upper=1.0),
            Interval(lower=-1.0, upper=0.0),
            NotImplementedError,
            id="invalid_bounds_and_thresholds_nonoverlapping",
        ),
    ],
)
def test_parameter_activation(
    bounds: Interval | None,
    thresholds: Interval,
    expected_result: Interval | type[ValueError] | type[NotImplementedError],
    mirror: bool,
):
    """The parameter activation utility correctly activates a parameter.

    Args:
        bounds: The bounds of the parameter to activate.
        thresholds: The inactivity thresholds.
        expected_result: The expected bounds of the activated parameter or any error
            raised.
        mirror: If ``True``, both bounds and thresholds get mirrored.
    """
    is_valid = isinstance(expected_result, Interval)

    if mirror:
        thresholds = mirror_interval(thresholds)
        if bounds is not None:
            bounds = mirror_interval(bounds)
        if is_valid:
            expected_result = mirror_interval(expected_result)

    if bounds is None:
        parameter = Mock()
    else:
        parameter = NumericalContinuousParameter("parameter", bounds=bounds)

    if is_valid:
        activated_parameter = activate_parameter(parameter, thresholds)
        assert activated_parameter.bounds == expected_result
    else:
        error_message = (
            "The thresholds must cover zero"
            if expected_result is ValueError
            else "proper sub-interval"
        )
        with pytest.raises(expected_result, match=error_message):
            activate_parameter(parameter, thresholds)
